# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Dataset Builder

A Dataset Builder is a builder class for generating a dataset by providing conditions.
"""
from __future__ import absolute_import

import datetime
from enum import Enum
import os
from typing import Any, Dict, List, Tuple, Union

import attr
import pandas as pd

from sagemaker import Session, s3, utils
from sagemaker.feature_store.feature_group import FeatureDefinition, FeatureGroup, FeatureTypeEnum


_DEFAULT_CATALOG = "AwsDataCatalog"
_DEFAULT_DATABASE = "sagemaker_featurestore"


@attr.s
class TableType(Enum):
    """Enum of Table types.

    The data type of a table can be FeatureGroup or DataFrame.
    """

    FEATURE_GROUP = "FeatureGroup"
    DATA_FRAME = "DataFrame"


@attr.s
class FeatureGroupToBeMerged:
    """FeatureGroup metadata which will be used for SQL join.

    This class instantiates a FeatureGroupToBeMerged object that comprises a list of feature names,
    a list of feature names which will be included in SQL query, a database, an Athena table name,
    a feature name of record identifier, a feature name of event time identifier and a feature name
    of base which is the target join key.

    Attributes:
        features (List[str]): A list of strings representing feature names of this FeatureGroup.
        included_feature_names (List[str]): A list of strings representing features to be
            included in the sql join.
        projected_feature_names (List[str]): A list of strings representing features to be
            included for final projection in output.
        catalog (str): A string representing the catalog.
        database (str): A string representing the database.
        table_name (str): A string representing the Athena table name of this FeatureGroup.
        record_dentifier_feature_name (str): A string representing the record identifier feature.
        event_time_identifier_feature (FeatureDefinition): A FeatureDefinition representing the
            event time identifier feature.
        target_feature_name_in_base (str): A string representing the feature name in base which will
            be used as target join key (default: None).
        table_type (TableType): A TableType representing the type of table if it is Feature Group or
            Panda Data Frame (default: None).
    """

    features: List[str] = attr.ib()
    included_feature_names: List[str] = attr.ib()
    projected_feature_names: List[str] = attr.ib()
    catalog: str = attr.ib()
    database: str = attr.ib()
    table_name: str = attr.ib()
    record_identifier_feature_name: str = attr.ib()
    event_time_identifier_feature: FeatureDefinition = attr.ib()
    target_feature_name_in_base: str = attr.ib(default=None)
    table_type: TableType = attr.ib(default=None)


def construct_feature_group_to_be_merged(
    feature_group: FeatureGroup,
    included_feature_names: List[str],
    target_feature_name_in_base: str = None,
) -> FeatureGroupToBeMerged:
    """Construct a FeatureGroupToBeMerged object by provided parameters.

    Args:
        feature_group (FeatureGroup): A FeatureGroup object.
        included_feature_names (List[str]): A list of strings representing features to be
            included in the output.
        target_feature_name_in_base (str): A string representing the feature name in base which
            will be used as target join key (default: None).
    Returns:
        A FeatureGroupToBeMerged object.

    Raises:
        ValueError: Invalid feature name(s) in included_feature_names.
    """
    feature_group_metadata = feature_group.describe()
    data_catalog_config = feature_group_metadata.get("OfflineStoreConfig", {}).get(
        "DataCatalogConfig", None
    )
    if not data_catalog_config:
        raise RuntimeError(f"No metastore is configured with FeatureGroup {feature_group.name}.")

    record_identifier_feature_name = feature_group_metadata.get("RecordIdentifierFeatureName", None)
    feature_definitions = feature_group_metadata.get("FeatureDefinitions", [])
    event_time_identifier_feature_name = feature_group_metadata.get("EventTimeFeatureName", None)
    event_time_identifier_feature_type = FeatureTypeEnum(
        next(
            filter(
                lambda f: f.get("FeatureName", None) == event_time_identifier_feature_name,
                feature_definitions,
            ),
            {},
        ).get("FeatureType", None)
    )
    table_name = data_catalog_config.get("TableName", None)
    database = data_catalog_config.get("Database", None)
    disable_glue = feature_group_metadata.get("DisableGlueTableCreation", False)
    catalog = data_catalog_config.get("Catalog", None) if disable_glue else _DEFAULT_CATALOG
    features = [feature.get("FeatureName", None) for feature in feature_definitions]

    for included_feature in included_feature_names or []:
        if included_feature not in features:
            raise ValueError(
                f"Feature {included_feature} not found in FeatureGroup {feature_group.name}"
            )
    if not included_feature_names:
        included_feature_names = features
        projected_feature_names = features.copy()
    else:
        projected_feature_names = included_feature_names.copy()
        if record_identifier_feature_name not in included_feature_names:
            included_feature_names.append(record_identifier_feature_name)
        if event_time_identifier_feature_name not in included_feature_names:
            included_feature_names.append(event_time_identifier_feature_name)
    return FeatureGroupToBeMerged(
        features,
        included_feature_names,
        projected_feature_names,
        catalog,
        database,
        table_name,
        record_identifier_feature_name,
        FeatureDefinition(event_time_identifier_feature_name, event_time_identifier_feature_type),
        target_feature_name_in_base,
        TableType.FEATURE_GROUP,
    )


@attr.s
class DatasetBuilder:
    """DatasetBuilder definition.

    This class instantiates a DatasetBuilder object that comprises a base, a list of feature names,
    an output path and a KMS key ID.

    Attributes:
        _sagemaker_session (Session): Session instance to perform boto calls.
        _base (Union[FeatureGroup, DataFrame]): A base which can be either a FeatureGroup or a
            pandas.DataFrame and will be used to merge other FeatureGroups and generate a Dataset.
        _output_path (str): An S3 URI which stores the output .csv file.
        _record_identifier_feature_name (str): A string representing the record identifier feature
            if base is a DataFrame (default: None).
        _event_time_identifier_feature_name (str): A string representing the event time identifier
            feature if base is a DataFrame (default: None).
        _included_feature_names (List[str]): A list of strings representing features to be
            included in the output (default: None).
        _kms_key_id (str): An KMS key id. If set, will be used to encrypt the result file
            (default: None).
        _point_in_time_accurate_join (bool): A boolean representing whether using point in time join
            or not (default: False).
        _include_duplicated_records (bool): A boolean representing whether including duplicated
            records or not (default: False).
        _include_deleted_records (bool): A boolean representing whether including deleted records or
            not (default: False).
        _number_of_recent_records (int): An int that how many records will be returned for each
            record identifier (default: 1).
        _number_of_records (int): An int that how many records will be returned (default: None).
        _write_time_ending_timestamp (datetime.datetime): A datetime that all records' write time in
            dataset will be before it (default: None).
        _event_time_starting_timestamp (datetime.datetime): A datetime that all records' event time
            in dataset will be after it (default: None).
        _event_time_ending_timestamp (datetime.datetime): A datetime that all records' event time in
            dataset will be before it (default: None).
        _feature_groups_to_be_merged (List[FeatureGroupToBeMerged]): A list of
            FeatureGroupToBeMerged which will be joined to base (default: []).
        _event_time_identifier_feature_type (FeatureTypeEnum): A FeatureTypeEnum representing the
            type of event time identifier feature (default: None).
    """

    _sagemaker_session: Session = attr.ib()
    _base: Union[FeatureGroup, pd.DataFrame] = attr.ib()
    _output_path: str = attr.ib()
    _record_identifier_feature_name: str = attr.ib(default=None)
    _event_time_identifier_feature_name: str = attr.ib(default=None)
    _included_feature_names: List[str] = attr.ib(default=None)
    _kms_key_id: str = attr.ib(default=None)

    _point_in_time_accurate_join: bool = attr.ib(init=False, default=False)
    _include_duplicated_records: bool = attr.ib(init=False, default=False)
    _include_deleted_records: bool = attr.ib(init=False, default=False)
    _number_of_recent_records: int = attr.ib(init=False, default=None)
    _number_of_records: int = attr.ib(init=False, default=None)
    _write_time_ending_timestamp: datetime.datetime = attr.ib(init=False, default=None)
    _event_time_starting_timestamp: datetime.datetime = attr.ib(init=False, default=None)
    _event_time_ending_timestamp: datetime.datetime = attr.ib(init=False, default=None)
    _feature_groups_to_be_merged: List[FeatureGroupToBeMerged] = attr.ib(init=False, factory=list)
    _event_time_identifier_feature_type: FeatureTypeEnum = attr.ib(default=None)

    _DATAFRAME_TYPE_TO_COLUMN_TYPE_MAP = {
        "object": "STRING",
        "int64": "INT",
        "float64": "DOUBLE",
        "bool": "BOOLEAN",
        "datetime64[ns]": "TIMESTAMP",
    }

    def with_feature_group(
        self,
        feature_group: FeatureGroup,
        target_feature_name_in_base: str = None,
        included_feature_names: List[str] = None,
    ):
        """Join FeatureGroup with base.

        Args:
            feature_group (FeatureGroup): A FeatureGroup which will be joined to base.
            target_feature_name_in_base (str): A string representing the feature name in base which
                will be used as target join key (default: None).
            included_feature_names (List[str]): A list of strings representing features to be
                included in the output (default: None).
        Returns:
            This DatasetBuilder object.
        """
        self._feature_groups_to_be_merged.append(
            construct_feature_group_to_be_merged(
                feature_group, included_feature_names, target_feature_name_in_base
            )
        )
        return self

    def point_in_time_accurate_join(self):
        """Set join type as point in time accurate join.

        Returns:
            This DatasetBuilder object.
        """
        self._point_in_time_accurate_join = True
        return self

    def include_duplicated_records(self):
        """Include duplicated records in dataset.

        Returns:
            This DatasetBuilder object.
        """
        self._include_duplicated_records = True
        return self

    def include_deleted_records(self):
        """Include deleted records in dataset.

        Returns:
            This DatasetBuilder object.
        """
        self._include_deleted_records = True
        return self

    def with_number_of_recent_records_by_record_identifier(self, number_of_recent_records: int):
        """Set number_of_recent_records field with provided input.

        Args:
            number_of_recent_records (int): An int that how many recent records will be returned for
                each record identifier.
        Returns:
            This DatasetBuilder object.
        """
        self._number_of_recent_records = number_of_recent_records
        return self

    def with_number_of_records_from_query_results(self, number_of_records: int):
        """Set number_of_records field with provided input.

        Args:
            number_of_records (int): An int that how many records will be returned.
        Returns:
            This DatasetBuilder object.
        """
        self._number_of_records = number_of_records
        return self

    def as_of(self, timestamp: datetime.datetime):
        """Set write_time_ending_timestamp field with provided input.

        Args:
            timestamp (datetime.datetime): A datetime that all records' write time in dataset will
                be before it.
        Returns:
            This DatasetBuilder object.
        """
        self._write_time_ending_timestamp = timestamp
        return self

    def with_event_time_range(
        self,
        starting_timestamp: datetime.datetime = None,
        ending_timestamp: datetime.datetime = None,
    ):
        """Set event_time_starting_timestamp and event_time_ending_timestamp with provided inputs.

        Args:
            starting_timestamp (datetime.datetime): A datetime that all records' event time in
                dataset will be after it (default: None).
            ending_timestamp (datetime.datetime): A datetime that all records' event time in dataset
                will be before it (default: None).
        Returns:
            This DatasetBuilder object.
        """
        self._event_time_starting_timestamp = starting_timestamp
        self._event_time_ending_timestamp = ending_timestamp
        return self

    def to_csv_file(self) -> Tuple[str, str]:
        """Get query string and result in .csv format file

        Returns:
            The S3 path of the .csv file.
            The query string executed.
        """
        if isinstance(self._base, pd.DataFrame):
            temp_id = utils.unique_name_from_base("dataframe-base")
            local_file_name = f"{temp_id}.csv"
            desired_s3_folder = f"{self._output_path}/{temp_id}"
            self._base.to_csv(local_file_name, index=False, header=False)
            s3.S3Uploader.upload(
                local_path=local_file_name,
                desired_s3_uri=desired_s3_folder,
                sagemaker_session=self._sagemaker_session,
                kms_key=self._kms_key_id,
            )
            os.remove(local_file_name)
            temp_table_name = f'dataframe_{temp_id.replace("-", "_")}'
            self._create_temp_table(temp_table_name, desired_s3_folder)
            base_features = list(self._base.columns)
            event_time_identifier_feature_dtype = self._base[
                self._event_time_identifier_feature_name
            ].dtypes
            self._event_time_identifier_feature_type = (
                FeatureGroup.DTYPE_TO_FEATURE_DEFINITION_CLS_MAP.get(
                    str(event_time_identifier_feature_dtype), None
                )
            )
            query_string = self._construct_query_string(
                FeatureGroupToBeMerged(
                    base_features,
                    self._included_feature_names if self._included_feature_names else base_features,
                    self._included_feature_names if self._included_feature_names else base_features,
                    _DEFAULT_CATALOG,
                    _DEFAULT_DATABASE,
                    temp_table_name,
                    self._record_identifier_feature_name,
                    FeatureDefinition(
                        self._event_time_identifier_feature_name,
                        self._event_time_identifier_feature_type,
                    ),
                    None,
                    TableType.DATA_FRAME,
                )
            )
            query_result = self._run_query(query_string, _DEFAULT_CATALOG, _DEFAULT_DATABASE)
            # TODO: cleanup temp table, need more clarification, keep it for now
            return query_result.get("QueryExecution", {}).get("ResultConfiguration", {}).get(
                "OutputLocation", None
            ), query_result.get("QueryExecution", {}).get("Query", None)
        if isinstance(self._base, FeatureGroup):
            base_feature_group = construct_feature_group_to_be_merged(
                self._base, self._included_feature_names
            )
            self._record_identifier_feature_name = base_feature_group.record_identifier_feature_name
            self._event_time_identifier_feature_name = (
                base_feature_group.event_time_identifier_feature.feature_name
            )
            self._event_time_identifier_feature_type = (
                base_feature_group.event_time_identifier_feature.feature_type
            )
            query_string = self._construct_query_string(base_feature_group)
            query_result = self._run_query(
                query_string,
                base_feature_group.catalog,
                base_feature_group.database,
            )
            return query_result.get("QueryExecution", {}).get("ResultConfiguration", {}).get(
                "OutputLocation", None
            ), query_result.get("QueryExecution", {}).get("Query", None)
        raise ValueError("Base must be either a FeatureGroup or a DataFrame.")

    def to_dataframe(self) -> Tuple[pd.DataFrame, str]:
        """Get query string and result in pandas.Dataframe

        Returns:
            The pandas.DataFrame object.
            The query string executed.
        """
        csv_file, query_string = self.to_csv_file()
        s3.S3Downloader.download(
            s3_uri=csv_file,
            local_path="./",
            kms_key=self._kms_key_id,
            sagemaker_session=self._sagemaker_session,
        )
        local_file_name = csv_file.split("/")[-1]
        df = pd.read_csv(local_file_name)
        os.remove(local_file_name)

        local_metadata_file_name = local_file_name + ".metadata"
        if os.path.exists(local_metadata_file_name):
            os.remove(local_file_name + ".metadata")

        if "row_recent" in df:
            df = df.drop("row_recent", axis="columns")
        return df, query_string

    def _construct_event_time_conditions(
        self,
        table_name: str,
        event_time_identifier_feature: FeatureDefinition,
    ) -> List[str]:
        """Internal method for constructing event time range sql range as string.

        Args:
            table_name (str): name of the table.
            event_time_identifier_feature (FeatureDefinition): A FeatureDefinition representing the
                event time identifier feature.
        Returns:
            The list of query strings.
        """
        event_time_conditions = []
        timestamp_cast_function_name = "from_unixtime"
        if event_time_identifier_feature.feature_type == FeatureTypeEnum.STRING:
            timestamp_cast_function_name = "from_iso8601_timestamp"
        if self._event_time_starting_timestamp:
            event_time_conditions.append(
                f"{timestamp_cast_function_name}({table_name}."
                + f'"{event_time_identifier_feature.feature_name}") >= '
                + f"from_unixtime({self._event_time_starting_timestamp.timestamp()})"
            )
        if self._event_time_ending_timestamp:
            event_time_conditions.append(
                f"{timestamp_cast_function_name}({table_name}."
                + f'"{event_time_identifier_feature.feature_name}") <= '
                + f"from_unixtime({self._event_time_ending_timestamp.timestamp()})"
            )
        return event_time_conditions

    def _construct_write_time_condition(
        self,
        table_name: str,
    ) -> str:
        """Internal method for constructing write time condition.

        Args:
            table_name (str): name of the table.
        Returns:
            string of write time condition.
        """
        write_time_condition = (
            f'{table_name}."write_time" <= '
            f"to_timestamp('{self._write_time_ending_timestamp.replace(microsecond=0)}', "
            f"'yyyy-mm-dd hh24:mi:ss')"
        )
        return write_time_condition

    def _construct_where_query_string(
        self,
        suffix: str,
        event_time_identifier_feature: FeatureDefinition,
        where_conditions: List[str],
    ) -> str:
        """Internal method for constructing SQL WHERE query string by parameters.

        Args:
            suffix (str): A temp identifier of the FeatureGroup.
            event_time_identifier_feature (FeatureDefinition): A FeatureDefinition representing the
                event time identifier feature.
            where_conditions (List[str]): A list of strings representing existing where clauses.
        Returns:
            The WHERE query string.

        Raises:
            ValueError: FeatureGroup not provided while using as_of(). Only found pandas.DataFrame.
        """
        if self._number_of_recent_records:
            if self._number_of_recent_records < 0:
                raise ValueError(
                    "Please provide non-negative integer for number_of_recent_records."
                )
        if self._number_of_records:
            if self._number_of_records < 0:
                raise ValueError("Please provide non-negative integer for number_of_records.")
        if self._include_deleted_records:
            if isinstance(self._base, pd.DataFrame):
                if len(self._feature_groups_to_be_merged) == 0:
                    raise ValueError(
                        "include_deleted_records() only works for FeatureGroup,"
                        " if there is no join operation."
                    )
        if self._include_duplicated_records:
            if isinstance(self._base, pd.DataFrame):
                if len(self._feature_groups_to_be_merged) == 0:
                    raise ValueError(
                        "include_duplicated_records() only works for FeatureGroup,"
                        " if there is no join operation."
                    )
        if self._point_in_time_accurate_join:
            if len(self._feature_groups_to_be_merged) == 0:
                raise ValueError(
                    "point_in_time_accurate_join() this operation only works when there is "
                    "more than one feature group to join."
                )
        if self._write_time_ending_timestamp:
            if isinstance(self._base, pd.DataFrame):
                if len(self._feature_groups_to_be_merged) == 0:
                    raise ValueError(
                        "as_of() only works for FeatureGroup," " if there is no join operation."
                    )
            if isinstance(self._base, FeatureGroup):
                if self._write_time_ending_timestamp:
                    where_conditions.append(self._construct_write_time_condition(f"table_{suffix}"))

        event_time_conditions = self._construct_event_time_conditions(
            f"table_{suffix}", event_time_identifier_feature
        )
        where_conditions.extend(event_time_conditions)

        if len(where_conditions) == 0:
            return ""
        return "WHERE " + "\nAND ".join(where_conditions)

    def _construct_dedup_query(self, feature_group: FeatureGroupToBeMerged, suffix: str) -> str:
        """Internal method for constructing removing duplicate records SQL query string.

        Args:
            feature_group (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object which has the
                FeatureGroup metadata.
            suffix (str): A temp identifier of the FeatureGroup.
        Returns:
            The SQL query string.
        """
        record_feature_name = feature_group.record_identifier_feature_name
        event_time_identifier_feature = feature_group.event_time_identifier_feature
        event_time_feature_name = feature_group.event_time_identifier_feature.feature_name
        rank_query_string = ""
        where_conditions = []
        where_conditions_str = ""
        is_dedup_enabled = False

        if feature_group.table_type is TableType.FEATURE_GROUP:
            is_dedup_enabled = True
            rank_query_string = (
                f'ORDER BY origin_{suffix}."api_invocation_time" DESC, '
                + f'origin_{suffix}."write_time" DESC\n'
            )

            if self._write_time_ending_timestamp:
                where_conditions.append(self._construct_write_time_condition(f"origin_{suffix}"))

        event_time_conditions = self._construct_event_time_conditions(
            f"origin_{suffix}", event_time_identifier_feature
        )
        where_conditions.extend(event_time_conditions)

        if len(where_conditions) != 0:
            where_conditions_str = "WHERE " + "\nAND ".join(where_conditions) + "\n"

        dedup_where_clause = f"WHERE dedup_row_{suffix} = 1\n" if is_dedup_enabled else ""
        return (
            f"table_{suffix} AS (\n"
            + "SELECT *\n"
            + "FROM (\n"
            + "SELECT *, row_number() OVER (\n"
            + f'PARTITION BY origin_{suffix}."{record_feature_name}", '
            + f'origin_{suffix}."{event_time_feature_name}"\n'
            + rank_query_string
            + f") AS dedup_row_{suffix}\n"
            + f'FROM "{feature_group.database}"."{feature_group.table_name}" origin_{suffix}\n'
            + where_conditions_str
            + ")\n"
            + dedup_where_clause
            + ")"
        )

    def _construct_deleted_query(self, feature_group: FeatureGroupToBeMerged, suffix: str) -> str:
        """Internal method for constructing removing deleted records SQL query string.

        Args:
            feature_group (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object which has the
                FeatureGroup metadata.
            suffix (str): A temp identifier of the FeatureGroup.
        Returns:
            The SQL query string.
        """
        record_feature_name = feature_group.record_identifier_feature_name
        event_time_identifier_feature = feature_group.event_time_identifier_feature
        event_time_feature_name = feature_group.event_time_identifier_feature.feature_name
        rank_query_string = f'ORDER BY origin_{suffix}."{event_time_feature_name}" DESC'
        write_time_condition = "\n"
        event_time_starting_condition = ""
        event_time_ending_condition = ""

        if feature_group.table_type is TableType.FEATURE_GROUP:
            rank_query_string += (
                f', origin_{suffix}."api_invocation_time" DESC, '
                + f'origin_{suffix}."write_time" DESC\n'
            )

            if self._write_time_ending_timestamp:
                write_time_condition += " AND "
                write_time_condition += self._construct_write_time_condition(f"origin_{suffix}")
                write_time_condition += "\n"

        if self._event_time_starting_timestamp and self._event_time_ending_timestamp:
            event_time_conditions = self._construct_event_time_conditions(
                f"origin_{suffix}", event_time_identifier_feature
            )
            event_time_starting_condition = "AND " + event_time_conditions[0] + "\n"
            event_time_ending_condition = "AND " + event_time_conditions[1] + "\n"

        return (
            f"deleted_{suffix} AS (\n"
            + "SELECT *\n"
            + "FROM (\n"
            + "SELECT *, row_number() OVER (\n"
            + f'PARTITION BY origin_{suffix}."{record_feature_name}"\n'
            + rank_query_string
            + f") AS deleted_row_{suffix}\n"
            + f'FROM "{feature_group.database}"."{feature_group.table_name}" origin_{suffix}\n'
            + "WHERE is_deleted"
            + write_time_condition
            + event_time_starting_condition
            + event_time_ending_condition
            + ")\n"
            + f"WHERE deleted_row_{suffix} = 1\n"
            + ")"
        )

    def _construct_table_included_features(
        self, feature_group: FeatureGroupToBeMerged, suffix: str
    ) -> str:
        """Internal method for constructing included features string of table.

        Args:
            feature_group (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object
                which has the metadata.
            suffix (str): A temp identifier of the table.
        Returns:
            The string that includes all feature to be included of table.
        """

        included_features = ", ".join(
            [
                f'table_{suffix}."{include_feature_name}"'
                for include_feature_name in feature_group.included_feature_names
            ]
        )
        return included_features

    def _construct_table_query(self, feature_group: FeatureGroupToBeMerged, suffix: str) -> str:
        """Internal method for constructing SQL query string by parameters.

        Args:
            feature_group (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object which has the
                FeatureGroup metadata.
            suffix (str): A temp identifier of the FeatureGroup.
        Returns:
            The query string.
        """
        included_features = self._construct_table_included_features(feature_group, suffix)

        # If base is a FeatureGroup then included_features_write_time will have a write_time column
        # Or included_features_write_time is same as included_features
        included_features_write_time = included_features

        if feature_group.table_type is TableType.FEATURE_GROUP:
            included_features_write_time += f', table_{suffix}."write_time"'
        record_feature_name = feature_group.record_identifier_feature_name
        event_time_feature_name = feature_group.event_time_identifier_feature.feature_name
        if self._include_duplicated_records and self._include_deleted_records:
            return (
                f"SELECT {included_features}\n"
                + f'FROM "{feature_group.database}"."{feature_group.table_name}" table_{suffix}\n'
                + self._construct_where_query_string(
                    suffix, feature_group.event_time_identifier_feature, ["NOT is_deleted"]
                )
            )
        if feature_group.table_type is TableType.FEATURE_GROUP and self._include_deleted_records:
            rank_query_string = ""
            if feature_group.table_type is TableType.FEATURE_GROUP:
                rank_query_string = (
                    f'ORDER BY origin_{suffix}."api_invocation_time" DESC, '
                    + f'origin_{suffix}."write_time" DESC\n'
                )
            return (
                f"SELECT {included_features}\n"
                + "FROM (\n"
                + "SELECT *, row_number() OVER (\n"
                + f'PARTITION BY origin_{suffix}."{record_feature_name}", '
                + f'origin_{suffix}."{event_time_feature_name}"\n'
                + rank_query_string
                + f") AS row_{suffix}\n"
                + f'FROM "{feature_group.database}"."{feature_group.table_name}" origin_{suffix}\n'
                + "WHERE NOT is_deleted"
                + f") AS table_{suffix}\n"
                + self._construct_where_query_string(
                    suffix,
                    feature_group.event_time_identifier_feature,
                    [f"row_{suffix} = 1"],
                )
            )
        rank_query_string = ""
        if feature_group.table_type is TableType.FEATURE_GROUP:
            rank_query_string = (
                f'OR (table_{suffix}."{event_time_feature_name}" = '
                + f'deleted_{suffix}."{event_time_feature_name}" '
                + f'AND table_{suffix}."api_invocation_time" > '
                + f'deleted_{suffix}."api_invocation_time")\n'
                + f'OR (table_{suffix}."{event_time_feature_name}" = '
                + f'deleted_{suffix}."{event_time_feature_name}" '
                + f'AND table_{suffix}."api_invocation_time" = '
                + f'deleted_{suffix}."api_invocation_time" '
                + f'AND table_{suffix}."write_time" > deleted_{suffix}."write_time")\n'
            )

        final_query_string = ""
        if feature_group.table_type is TableType.FEATURE_GROUP:
            if self._include_duplicated_records:
                final_query_string = (
                    f"WITH {self._construct_deleted_query(feature_group, suffix)}\n"
                    + f"SELECT {included_features}\n"
                    + "FROM (\n"
                    + f"SELECT {included_features_write_time}\n"
                    + f'FROM "{feature_group.database}"."{feature_group.table_name}"'
                    + f" table_{suffix}\n"
                    + f"LEFT JOIN deleted_{suffix}\n"
                    + f'ON table_{suffix}."{record_feature_name}" = '
                    + f'deleted_{suffix}."{record_feature_name}"\n'
                    + f'WHERE deleted_{suffix}."{record_feature_name}" IS NULL\n'
                    + "UNION ALL\n"
                    + f"SELECT {included_features_write_time}\n"
                    + f"FROM deleted_{suffix}\n"
                    + f'JOIN "{feature_group.database}"."{feature_group.table_name}"'
                    + f" table_{suffix}\n"
                    + f'ON table_{suffix}."{record_feature_name}" = '
                    + f'deleted_{suffix}."{record_feature_name}"\n'
                    + "AND (\n"
                    + f'table_{suffix}."{event_time_feature_name}" > '
                    + f'deleted_{suffix}."{event_time_feature_name}"\n'
                    + rank_query_string
                    + ")\n"
                    + f") AS table_{suffix}\n"
                    + self._construct_where_query_string(
                        suffix, feature_group.event_time_identifier_feature, []
                    )
                )
            else:
                final_query_string = (
                    f"WITH {self._construct_dedup_query(feature_group, suffix)},\n"
                    + f"{self._construct_deleted_query(feature_group, suffix)}\n"
                    + f"SELECT {included_features}\n"
                    + "FROM (\n"
                    + f"SELECT {included_features_write_time}\n"
                    + f"FROM table_{suffix}\n"
                    + f"LEFT JOIN deleted_{suffix}\n"
                    + f'ON table_{suffix}."{record_feature_name}" = '
                    + f'deleted_{suffix}."{record_feature_name}"\n'
                    + f'WHERE deleted_{suffix}."{record_feature_name}" IS NULL\n'
                    + "UNION ALL\n"
                    + f"SELECT {included_features_write_time}\n"
                    + f"FROM deleted_{suffix}\n"
                    + f"JOIN table_{suffix}\n"
                    + f'ON table_{suffix}."{record_feature_name}" = '
                    + f'deleted_{suffix}."{record_feature_name}"\n'
                    + "AND (\n"
                    + f'table_{suffix}."{event_time_feature_name}" > '
                    + f'deleted_{suffix}."{event_time_feature_name}"\n'
                    + rank_query_string
                    + ")\n"
                    + f") AS table_{suffix}\n"
                    + self._construct_where_query_string(
                        suffix, feature_group.event_time_identifier_feature, []
                    )
                )
        else:
            final_query_string = (
                f"WITH {self._construct_dedup_query(feature_group, suffix)}\n"
                + f"SELECT {included_features}\n"
                + "FROM (\n"
                + f"SELECT {included_features_write_time}\n"
                + f"FROM table_{suffix}\n"
                + f") AS table_{suffix}\n"
                + self._construct_where_query_string(
                    suffix, feature_group.event_time_identifier_feature, []
                )
            )
        return final_query_string

    def _construct_query_string(self, base: FeatureGroupToBeMerged) -> str:
        """Internal method for constructing SQL query string by parameters.

        Args:
            base (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object which has the metadata.
        Returns:
            The query string.

        Raises:
            ValueError: target_feature_name_in_base is an invalid feature name.
        """
        base_table_query_string = self._construct_table_query(base, "base")
        query_string = f"WITH fg_base AS ({base_table_query_string})"
        if len(self._feature_groups_to_be_merged) > 0:
            with_subquery_string = "".join(
                [
                    f",\nfg_{i} AS ({self._construct_table_query(feature_group, str(i))})"
                    for i, feature_group in enumerate(self._feature_groups_to_be_merged)
                ]
            )
            query_string += with_subquery_string

        selected_features = ""
        selected_features += ", ".join(map("fg_base.{0}".format, base.projected_feature_names))
        if len(self._feature_groups_to_be_merged) > 0:
            for i, feature_group in enumerate(self._feature_groups_to_be_merged):
                selected_features += ", "
                selected_features += ", ".join(
                    [
                        f'fg_{i}."{feature_name}" as "{feature_name}.{(i+1)}"'
                        for feature_name in feature_group.projected_feature_names
                    ]
                )

        selected_features_final = ""
        selected_features_final += ", ".join(base.projected_feature_names)
        if len(self._feature_groups_to_be_merged) > 0:
            for i, feature_group in enumerate(self._feature_groups_to_be_merged):
                selected_features_final += ", "
                selected_features_final += ", ".join(
                    [
                        '"{0}.{1}"'.format(feature_name, (i + 1))
                        for feature_name in feature_group.projected_feature_names
                    ]
                )

        query_string += (
            f"\nSELECT {selected_features_final}\n"
            + "FROM (\n"
            + f"SELECT {selected_features}, row_number() OVER (\n"
            + f'PARTITION BY fg_base."{base.record_identifier_feature_name}"\n'
            + f'ORDER BY fg_base."{base.event_time_identifier_feature.feature_name}" DESC'
        )

        recent_record_where_clause = ""
        if self._number_of_recent_records is not None and self._number_of_recent_records >= 0:
            recent_record_where_clause = f"WHERE row_recent <= {self._number_of_recent_records}"

        join_subquery_strings = []
        if len(self._feature_groups_to_be_merged) > 0:
            for i, feature_group in enumerate(self._feature_groups_to_be_merged):
                if not feature_group.target_feature_name_in_base:
                    feature_group.target_feature_name_in_base = self._record_identifier_feature_name
                else:
                    if feature_group.target_feature_name_in_base not in base.features:
                        raise ValueError(
                            f"Feature {feature_group.target_feature_name_in_base} not found in base"
                        )
                query_string += (
                    f', fg_{i}."{feature_group.event_time_identifier_feature.feature_name}" DESC'
                )
                join_subquery_strings.append(self._construct_join_condition(feature_group, str(i)))

        query_string += (
            "\n) AS row_recent\n"
            + "FROM fg_base"
            + "".join(join_subquery_strings)
            + "\n)\n"
            + f"{recent_record_where_clause}"
        )

        if self._number_of_records is not None and self._number_of_records >= 0:
            query_string += f"\nLIMIT {self._number_of_records}"
        return query_string

    def _construct_join_condition(self, feature_group: FeatureGroupToBeMerged, suffix: str) -> str:
        """Internal method for constructing SQL JOIN query string by parameters.

        Args:
            feature_group (FeatureGroupToBeMerged): A FeatureGroupToBeMerged object which has the
                FeatureGroup metadata.
            suffix (str): A temp identifier of the FeatureGroup.
        Returns:
            The JOIN query string.
        """
        join_condition_string = (
            f"\nJOIN fg_{suffix}\n"
            + f'ON fg_base."{feature_group.target_feature_name_in_base}" = '
            + f'fg_{suffix}."{feature_group.record_identifier_feature_name}"'
        )
        base_timestamp_cast_function_name = "from_unixtime"
        if self._event_time_identifier_feature_type == FeatureTypeEnum.STRING:
            base_timestamp_cast_function_name = "from_iso8601_timestamp"
        timestamp_cast_function_name = "from_unixtime"
        if feature_group.event_time_identifier_feature.feature_type == FeatureTypeEnum.STRING:
            timestamp_cast_function_name = "from_iso8601_timestamp"
        if self._point_in_time_accurate_join:
            join_condition_string += (
                f"\nAND {base_timestamp_cast_function_name}(fg_base."
                + f'"{self._event_time_identifier_feature_name}") >= '
                + f"{timestamp_cast_function_name}(fg_{suffix}."
                + f'"{feature_group.event_time_identifier_feature.feature_name}")'
            )
        return join_condition_string

    def _create_temp_table(self, temp_table_name: str, desired_s3_folder: str):
        """Internal method for creating a temp Athena table for the base pandas.Dataframe.

        Args:
            temp_table_name (str): The Athena table name of base pandas.DataFrame.
            desired_s3_folder (str): The S3 URI of the folder of the data.
        """
        columns_string = ", ".join(
            [self._construct_athena_table_column_string(column) for column in self._base.columns]
        )
        serde_properties = '"separatorChar" = ",", "quoteChar" = "`", "escapeChar" = "\\\\"'
        query_string = (
            f"CREATE EXTERNAL TABLE {temp_table_name} ({columns_string}) "
            + "ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.OpenCSVSerde' "
            + f"WITH SERDEPROPERTIES ({serde_properties}) "
            + f"LOCATION '{desired_s3_folder}';"
        )
        self._run_query(query_string, _DEFAULT_CATALOG, _DEFAULT_DATABASE)

    def _construct_athena_table_column_string(self, column: str) -> str:
        """Internal method for constructing string of Athena column.

        Args:
            column (str): The column name from pandas.Dataframe.
        Returns:
            The Athena column string.

        Raises:
            RuntimeError: The type of pandas.Dataframe column is not support yet.
        """
        dataframe_type = self._base[column].dtypes
        if str(dataframe_type) not in self._DATAFRAME_TYPE_TO_COLUMN_TYPE_MAP.keys():
            raise RuntimeError(f"The dataframe type {dataframe_type} is not supported yet.")
        return f"{column} {self._DATAFRAME_TYPE_TO_COLUMN_TYPE_MAP.get(str(dataframe_type), None)}"

    def _run_query(self, query_string: str, catalog: str, database: str) -> Dict[str, Any]:
        """Internal method for execute Athena query, wait for query finish and get query result.

        Args:
            query_string (str): The SQL query statements to be executed.
            catalog (str): The name of the data catalog used in the query execution.
            database (str): The name of the database used in the query execution.
        Returns:
            The query result.

        Raises:
            RuntimeError: Athena query failed.
        """
        query = self._sagemaker_session.start_query_execution(
            catalog=catalog,
            database=database,
            query_string=query_string,
            output_location=self._output_path,
            kms_key=self._kms_key_id,
        )
        query_id = query.get("QueryExecutionId", None)
        self._sagemaker_session.wait_for_athena_query(query_execution_id=query_id)
        query_result = self._sagemaker_session.get_query_execution(query_execution_id=query_id)
        query_state = query_result.get("QueryExecution", {}).get("Status", {}).get("State", None)

        if query_state != "SUCCEEDED":
            raise RuntimeError(f"Failed to execute query {query_id}.")
        return query_result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Schema constraints for model card attribute value."""
from __future__ import absolute_import

from enum import Enum


class ModelCardStatusEnum(str, Enum):
    """Model card status enumerator"""

    DRAFT = "Draft"
    PENDING_REVIEW = "PendingReview"
    APPROVED = "Approved"
    ARCHIVED = "Archived"


class RiskRatingEnum(str, Enum):
    """Risk rating enumerator"""

    HIGH = "High"
    MEDIUM = "Medium"
    LOW = "Low"
    UNKNOWN = "Unknown"


class ObjectiveFunctionEnum(str, Enum):
    """Objective function enumerator"""

    MAXIMIZE = "Maximize"
    MINIMIZE = "Minimize"


class FacetEnum(str, Enum):
    """Objective function facet enumerator"""

    LOSS = "Loss"
    ACCURACY = "Accuracy"
    RMSE = "RMSE"
    MAE = "MAE"
    AUC = "AUC"


class MetricTypeEnum(str, Enum):
    """Metric type enumerator"""

    NUMBER = "number"
    LINEAR_GRAPH = "linear_graph"
    STRING = "string"
    BOOLEAN = "boolean"
    MATRIX = "matrix"
    BAR_CHART = "bar_chart"


METRIC_VALUE_TYPE_MAP = {
    MetricTypeEnum.NUMBER: [int, float],
    MetricTypeEnum.LINEAR_GRAPH: [list],
    MetricTypeEnum.STRING: [str],
    MetricTypeEnum.BOOLEAN: [bool],
    MetricTypeEnum.MATRIX: [list],
    MetricTypeEnum.BAR_CHART: [list],
}


PYTHON_TYPE_TO_METRIC_VALUE_TYPE = {
    int: MetricTypeEnum.NUMBER,
    float: MetricTypeEnum.NUMBER,
    str: MetricTypeEnum.STRING,
    bool: MetricTypeEnum.BOOLEAN,
}

MODEL_ARTIFACT_MAX_SIZE = 15
ENVIRONMENT_CONTAINER_IMAGES_MAX_SIZE = 15
TRAINING_DATASETS_MAX_SIZE = 15
TRAINING_METRICS_MAX_SIZE = 50
USER_PROVIDED_TRAINING_METRICS_MAX_SIZE = 50
EVALUATION_DATASETS_MAX_SIZE = 10



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Utilities to support workflow."""
from __future__ import absolute_import

import inspect
import logging
from functools import wraps
from pathlib import Path
from typing import List, Sequence, Union, Set, TYPE_CHECKING
import hashlib
from urllib.parse import unquote, urlparse
from contextlib import contextmanager
from _hashlib import HASH as Hash

from sagemaker.workflow.parameters import Parameter
from sagemaker.workflow.pipeline_context import _StepArguments, _PipelineConfig
from sagemaker.workflow.entities import (
    Entity,
    RequestType,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from sagemaker.workflow.step_collections import StepCollection

BUF_SIZE = 65536  # 64KiB
_pipeline_config: _PipelineConfig = None


def list_to_request(entities: Sequence[Union[Entity, "StepCollection"]]) -> List[RequestType]:
    """Get the request structure for list of entities.

    Args:
        entities (Sequence[Entity]): A list of entities.
    Returns:
        list: A request structure for a workflow service call.
    """
    from sagemaker.workflow.step_collections import StepCollection

    request_dicts = []
    for entity in entities:
        if isinstance(entity, Entity):
            request_dicts.append(entity.to_request())
        elif isinstance(entity, StepCollection):
            request_dicts.extend(entity.request_dicts())
    return request_dicts


@contextmanager
def _pipeline_config_manager(pipeline_name: str, step_name: str, code_hash: str, config_hash: str):
    """Expose static _pipeline_config variable to other modules

    Args:
        pipeline_name (str): pipeline name
        step_name (str): step name
        code_hash (str): a hash of the code artifact for the particular step
        config_hash (str): a hash of the config artifact for the particular step (Processing)
    """

    # pylint: disable=W0603
    global _pipeline_config
    _pipeline_config = _PipelineConfig(pipeline_name, step_name, code_hash, config_hash)
    try:
        yield
    finally:
        _pipeline_config = None


def build_steps(steps: Sequence[Entity], pipeline_name: str):
    """Get the request structure for list of steps, with _pipeline_config_manager

    Args:
        steps (Sequence[Entity]): A list of steps, (Entity type because Step causes circular import)
        pipeline_name (str): The name of the pipeline, passed down from pipeline.to_request()
    Returns:
        list: A request structure object for a service call for the list of pipeline steps
    """
    from sagemaker.workflow.step_collections import StepCollection

    request_dicts = []
    for step in steps:
        if isinstance(step, StepCollection):
            request_dicts.extend(step.request_dicts())
        else:
            with _pipeline_config_manager(
                pipeline_name, step.name, get_code_hash(step), get_config_hash(step)
            ):
                request_dicts.append(step.to_request())
    return request_dicts


def get_code_hash(step: Entity) -> str:
    """Get the hash of the code artifact(s) for the given step

    Args:
        step (Entity): A pipeline step object (Entity type because Step causes circular import)
    Returns:
        str: A hash string representing the unique code artifact(s) for the step
    """
    from sagemaker.workflow.steps import ProcessingStep, TrainingStep

    if isinstance(step, ProcessingStep) and step.step_args:
        kwargs = step.step_args.func_kwargs
        source_dir = kwargs.get("source_dir")
        submit_class = kwargs.get("submit_class")
        dependencies = get_processing_dependencies(
            [
                kwargs.get("dependencies"),
                kwargs.get("submit_py_files"),
                [submit_class] if submit_class else None,
                kwargs.get("submit_jars"),
                kwargs.get("submit_files"),
            ]
        )
        code = kwargs.get("submit_app") or kwargs.get("code")

        return get_processing_code_hash(code, source_dir, dependencies)

    if isinstance(step, TrainingStep) and step.step_args:
        job_obj = step.step_args.func_args[0]
        source_dir = job_obj.source_dir
        dependencies = job_obj.dependencies
        entry_point = job_obj.entry_point

        return get_training_code_hash(entry_point, source_dir, dependencies)
    return None


def get_processing_dependencies(dependency_args: List[List[str]]) -> List[str]:
    """Get the Processing job dependencies from the processor run kwargs

    Args:
        dependency_args: A list of dependency args from processor.run()
    Returns:
        List[str]: A list of code dependencies for the job
    """

    dependencies = []
    for arg in dependency_args:
        if arg:
            dependencies += arg

    return dependencies


def get_processing_code_hash(code: str, source_dir: str, dependencies: List[str]) -> str:
    """Get the hash of a processing step's code artifact(s).

    Args:
        code (str): Path to a file with the processing script to run
        source_dir (str): Path to a directory with any other processing
                source code dependencies aside from the entry point file
        dependencies (str): A list of paths to directories (absolute
                or relative) with any additional libraries that will be exported
                to the container
    Returns:
        str: A hash string representing the unique code artifact(s) for the step
    """

    # If FrameworkProcessor contains source_dir
    if source_dir:
        source_dir_url = urlparse(source_dir)
        if source_dir_url.scheme == "" or source_dir_url.scheme == "file":
            return hash_files_or_dirs([source_dir] + dependencies)
    # Other Processors - Spark, Script, Base, etc.
    if code:
        code_url = urlparse(code)
        if code_url.scheme == "" or code_url.scheme == "file":
            return hash_files_or_dirs([code] + dependencies)
    return None


def get_training_code_hash(entry_point: str, source_dir: str, dependencies: List[str]) -> str:
    """Get the hash of a training step's code artifact(s).

    Args:
        entry_point (str): The absolute or relative path to the local Python
                source file that should be executed as the entry point to
                training
        source_dir (str): Path to a directory with any other training source
                code dependencies aside from the entry point file
        dependencies (str): A list of paths to directories (absolute
                or relative) with any additional libraries that will be exported
                to the container
    Returns:
        str: A hash string representing the unique code artifact(s) for the step
    """
    from sagemaker.workflow import is_pipeline_variable

    if not is_pipeline_variable(source_dir) and not is_pipeline_variable(entry_point):
        if source_dir:
            source_dir_url = urlparse(source_dir)
            if source_dir_url.scheme == "" or source_dir_url.scheme == "file":
                return hash_files_or_dirs([source_dir] + dependencies)
        elif entry_point:
            entry_point_url = urlparse(entry_point)
            if entry_point_url.scheme == "" or entry_point_url.scheme == "file":
                return hash_files_or_dirs([entry_point] + dependencies)
    return None


def get_config_hash(step: Entity):
    """Get the hash of the config artifact(s) for the given step

    Args:
        step (Entity): A pipeline step object (Entity type because Step causes circular import)
    Returns:
        str: A hash string representing the unique config artifact(s) for the step
    """
    from sagemaker.workflow.steps import ProcessingStep

    if isinstance(step, ProcessingStep) and step.step_args:
        config = step.step_args.func_kwargs.get("configuration")
        if config:
            return hash_object(config)
    return None


def hash_object(obj) -> str:
    """Get the MD5 hash of an object.

    Args:
        obj (dict): The object
    Returns:
        str: The MD5 hash of the object
    """
    return hashlib.md5(str(obj).encode()).hexdigest()


def hash_file(path: str) -> str:
    """Get the MD5 hash of a file.

    Args:
        path (str): The local path for the file.
    Returns:
        str: The MD5 hash of the file.
    """
    return _hash_file(path, hashlib.md5()).hexdigest()


def hash_files_or_dirs(paths: List[str]) -> str:
    """Get the MD5 hash of the contents of a list of files or directories.

    Hash is changed if:
       * input list is changed
       * new nested directories/files are added to any directory in the input list
       * nested directory/file names are changed for any of the inputted directories
       * content of files is edited

    Args:
        paths: List of file or directory paths
    Returns:
        str: The MD5 hash of the list of files or directories.
    """
    md5 = hashlib.md5()
    for path in sorted(paths):
        md5 = _hash_file_or_dir(path, md5)
    return md5.hexdigest()


def _hash_file_or_dir(path: str, md5: Hash) -> Hash:
    """Updates the inputted Hash with the contents of the current path.

    Args:
        path: path of file or directory
    Returns:
        str: The MD5 hash of the file or directory
    """
    if isinstance(path, str) and path.lower().startswith("file://"):
        path = unquote(urlparse(path).path)
    md5.update(path.encode())
    if Path(path).is_dir():
        md5 = _hash_dir(path, md5)
    elif Path(path).is_file():
        md5 = _hash_file(path, md5)
    return md5


def _hash_dir(directory: Union[str, Path], md5: Hash) -> Hash:
    """Updates the inputted Hash with the contents of the current path.

    Args:
        directory: path of the directory
    Returns:
        str: The MD5 hash of the directory
    """
    if not Path(directory).is_dir():
        raise ValueError(str(directory) + " is not a valid directory")
    for path in sorted(Path(directory).iterdir()):
        md5.update(path.name.encode())
        if path.is_file():
            md5 = _hash_file(path, md5)
        elif path.is_dir():
            md5 = _hash_dir(path, md5)
    return md5


def _hash_file(file: Union[str, Path], md5: Hash) -> Hash:
    """Updates the inputted Hash with the contents of the current path.

    Args:
        file: path of the file
    Returns:
        str: The MD5 hash of the file
    """
    if isinstance(file, str) and file.lower().startswith("file://"):
        file = unquote(urlparse(file).path)
    if not Path(file).is_file():
        raise ValueError(str(file) + " is not a valid file")
    with open(file, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5


def validate_step_args_input(
    step_args: _StepArguments, expected_caller: Set[str], error_message: str
):
    """Validate the `_StepArguments` object which is passed into a pipeline step

    Args:
        step_args (_StepArguments): A `_StepArguments` object to be used for composing
            a pipeline step.
        expected_caller (Set[str]): The expected name of the caller function which is
            intercepted by the PipelineSession to get the step arguments.
        error_message (str): The error message to be thrown if the validation fails.
    """
    if not isinstance(step_args, _StepArguments):
        raise TypeError(error_message)
    if step_args.caller_name not in expected_caller:
        raise ValueError(error_message)


def override_pipeline_parameter_var(func):
    """A decorator to override pipeline Parameters passed into a function

    This is a temporary decorator to override pipeline Parameter objects with their default value
    and display warning information to instruct users to update their code.

    This decorator can help to give a grace period for users to update their code when
    we make changes to explicitly prevent passing any pipeline variables to a function.

    We should remove this decorator after the grace period.
    """
    warning_msg_template = (
        "The input argument %s of function (%s) is a pipeline variable (%s), which is not allowed. "
        "The default_value of this Parameter object will be used to override it. "
        "Please make sure the default_value is valid."
    )

    @wraps(func)
    def wrapper(*args, **kwargs):
        func_name = "{}.{}".format(func.__module__, func.__name__)
        params = inspect.signature(func).parameters
        args = list(args)
        for i, (arg_name, _) in enumerate(params.items()):
            if i >= len(args):
                break
            if isinstance(args[i], Parameter):
                logger.warning(warning_msg_template, arg_name, func_name, type(args[i]))
                args[i] = args[i].default_value
        args = tuple(args)

        for arg_name, value in kwargs.items():
            if isinstance(value, Parameter):
                logger.warning(warning_msg_template, arg_name, func_name, type(value))
                kwargs[arg_name] = value.default_value
        return func(*args, **kwargs)

    return wrapper


def execute_job_functions(step_args: _StepArguments):
    """Execute the job class functions during pipeline definition construction

    Executes the job functions such as run(), fit(), or transform() that have been
    delayed until the pipeline gets built, for steps built with a PipelineSession.

    Handles multiple functions in instances where job functions are chained
    together from the inheritance of different job classes (e.g. PySparkProcessor,
    ScriptProcessor, and Processor).

    Args:
        step_args (_StepArguments): A `_StepArguments` object to be used for composing
            a pipeline step, contains the necessary function information
    """

    chained_args = step_args.func(*step_args.func_args, **step_args.func_kwargs)
    if isinstance(chained_args, _StepArguments):
        execute_job_functions(chained_args)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains helper methods related to Lambda."""
from __future__ import print_function, absolute_import

from io import BytesIO
import zipfile
import time
from botocore.exceptions import ClientError
from sagemaker.session import Session


class Lambda:
    """Contains lambda boto3 wrappers to Create, Update, Delete and Invoke Lambda functions."""

    def __init__(
        self,
        function_arn: str = None,
        function_name: str = None,
        execution_role_arn: str = None,
        zipped_code_dir: str = None,
        s3_bucket: str = None,
        script: str = None,
        handler: str = None,
        session: Session = None,
        timeout: int = 120,
        memory_size: int = 128,
        runtime: str = "python3.8",
    ):
        """Constructs a Lambda instance.

        This instance represents a Lambda function and provides methods for updating,
        deleting and invoking the function.

        This class can be used either for creating a new Lambda function or using an existing one.
        When using an existing Lambda function, only the function_arn argument is required.
        When creating a new one the function_name, execution_role_arn and handler arguments
        are required, as well as either script or zipped_code_dir.

        Args:
            function_arn (str): The arn of the Lambda function.
            function_name (str): The name of the Lambda function.
                Function name must be provided to create a Lambda function.
            execution_role_arn (str): The role to be attached to Lambda function.
            zipped_code_dir (str): The path of the zipped code package of the Lambda function.
            s3_bucket (str): The bucket where zipped code is uploaded.
                If not provided, default session bucket is used to upload zipped_code_dir.
            script (str): The path of Lambda function script for direct zipped upload
            handler (str): The Lambda handler. The format for handler should be
                file_name.function_name. For ex: if the name of the Lambda script is
                hello_world.py and Lambda function definition in that script is
                lambda_handler(event, context), the handler should be hello_world.lambda_handler
            session (sagemaker.session.Session): Session object which manages interactions
                with Amazon SageMaker APIs and any other AWS services needed.
                If not specified, new session is created.
            timeout (int): Timeout of the Lambda function in seconds. Default is 120 seconds.
            memory_size (int): Memory of the Lambda function in megabytes. Default is 128 MB.
            runtime (str): Runtime of the Lambda function. Default is set to python3.8.
        """
        self.function_arn = function_arn
        self.function_name = function_name
        self.zipped_code_dir = zipped_code_dir
        self.s3_bucket = s3_bucket
        self.script = script
        self.handler = handler
        self.execution_role_arn = execution_role_arn
        self.session = session if session is not None else Session()
        self.timeout = timeout
        self.memory_size = memory_size
        self.runtime = runtime

        if function_arn is None and function_name is None:
            raise ValueError("Either function_arn or function_name must be provided.")

        if function_name is not None:
            if execution_role_arn is None:
                raise ValueError("execution_role_arn must be provided.")
            if zipped_code_dir is None and script is None:
                raise ValueError("Either zipped_code_dir or script must be provided.")
            if zipped_code_dir and script:
                raise ValueError("Provide either script or zipped_code_dir, not both.")
            if handler is None:
                raise ValueError("Lambda handler must be provided.")

    def create(self):
        """Method to create a lambda function.

        Returns: boto3 response from Lambda's create_function method.
        """
        lambda_client = _get_lambda_client(self.session)

        if self.function_name is None:
            raise ValueError("FunctionName must be provided to create a Lambda function.")

        if self.script is not None:
            code = {"ZipFile": _zip_lambda_code(self.script)}
        else:
            bucket = self.s3_bucket or self.session.default_bucket()
            key = _upload_to_s3(
                s3_client=_get_s3_client(self.session),
                function_name=self.function_name,
                zipped_code_dir=self.zipped_code_dir,
                s3_bucket=bucket,
            )
            code = {"S3Bucket": bucket, "S3Key": key}

        try:
            response = lambda_client.create_function(
                FunctionName=self.function_name,
                Runtime=self.runtime,
                Handler=self.handler,
                Role=self.execution_role_arn,
                Code=code,
                Timeout=self.timeout,
                MemorySize=self.memory_size,
            )
            return response
        except ClientError as e:
            error = e.response["Error"]
            raise ValueError(error)

    def update(self):
        """Method to update a lambda function.

        Returns: boto3 response from Lambda's update_function method.
        """
        lambda_client = _get_lambda_client(self.session)
        retry_attempts = 7
        for i in range(retry_attempts):
            try:
                if self.script is not None:
                    response = lambda_client.update_function_code(
                        FunctionName=self.function_name, ZipFile=_zip_lambda_code(self.script)
                    )
                else:
                    response = lambda_client.update_function_code(
                        FunctionName=(self.function_name or self.function_arn),
                        S3Bucket=self.s3_bucket,
                        S3Key=_upload_to_s3(
                            s3_client=_get_s3_client(self.session),
                            function_name=self.function_name,
                            zipped_code_dir=self.zipped_code_dir,
                            s3_bucket=self.s3_bucket,
                        ),
                    )
                return response
            except ClientError as e:
                error = e.response["Error"]
                code = error["Code"]
                if code == "ResourceConflictException":
                    if i == retry_attempts - 1:
                        raise ValueError(error)
                    # max wait time = 2**0 + 2**1 + .. + 2**6 = 127 seconds
                    time.sleep(2**i)
                else:
                    raise ValueError(error)

    def upsert(self):
        """Method to create a lambda function or update it if it already exists

        Returns: boto3 response from Lambda's methods.
        """
        try:
            return self.create()
        except ValueError as error:
            if "ResourceConflictException" in str(error):
                return self.update()
            raise

    def invoke(self):
        """Method to invoke a lambda function.

        Returns: boto3 response from Lambda's invoke method.
        """
        lambda_client = _get_lambda_client(self.session)
        try:
            response = lambda_client.invoke(
                FunctionName=self.function_name or self.function_arn,
                InvocationType="RequestResponse",
            )
            return response
        except ClientError as e:
            error = e.response["Error"]
            raise ValueError(error)

    def delete(self):
        """Method to delete a lambda function.

        Returns: boto3 response from Lambda's delete_function method.
        """
        lambda_client = _get_lambda_client(self.session)
        try:
            response = lambda_client.delete_function(
                FunctionName=self.function_name or self.function_arn
            )
            return response
        except ClientError as e:
            error = e.response["Error"]
            raise ValueError(error)


def _get_s3_client(session):
    """Method to get a boto3 s3 client.

    Returns: a s3 client.
    """
    sagemaker_session = session or Session()
    if sagemaker_session.s3_client is None:
        s3_client = sagemaker_session.boto_session.client(
            "s3", region_name=sagemaker_session.boto_region_name
        )
    else:
        s3_client = sagemaker_session.s3_client
    return s3_client


def _get_lambda_client(session):
    """Method to get a boto3 lambda client.

    Returns: a lambda client.
    """
    sagemaker_session = session or Session()
    if sagemaker_session.lambda_client is None:
        lambda_client = sagemaker_session.boto_session.client(
            "lambda", region_name=sagemaker_session.boto_region_name
        )
    else:
        lambda_client = sagemaker_session.lambda_client
    return lambda_client


def _upload_to_s3(s3_client, function_name, zipped_code_dir, s3_bucket):
    """Upload the zipped code to S3 bucket provided in the Lambda instance.

    Lambda instance must have a path to the zipped code folder and a S3 bucket to upload
    the code. The key will lambda/function_name/code and the S3 URI where the code is
    uploaded is in this format: s3://bucket_name/lambda/function_name/code.

    Returns: the S3 key where the code is uploaded.
    """
    key = "{}/{}/{}".format("lambda", function_name, "code")
    s3_client.upload_file(zipped_code_dir, s3_bucket, key)
    return key


def _zip_lambda_code(script):
    """This method zips the lambda function script.

    Lambda function script is provided in the lambda instance and reads that zipped file.

    Returns: A buffer of zipped lambda function script.
    """
    buffer = BytesIO()
    code_dir = script.split("/")[-1]

    with zipfile.ZipFile(buffer, "w") as z:
        z.write(script, code_dir)
    buffer.seek(0)
    return buffer.read()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains code related to HuggingFace Processors which are used for Processing jobs.

These jobs let customers perform data pre-processing, post-processing, feature engineering,
data validation, and model evaluation and interpretation on SageMaker.
"""
from __future__ import absolute_import

from typing import Union, Optional, List, Dict

from sagemaker.session import Session
from sagemaker.network import NetworkConfig
from sagemaker.processing import FrameworkProcessor
from sagemaker.huggingface.estimator import HuggingFace

from sagemaker.workflow.entities import PipelineVariable


class HuggingFaceProcessor(FrameworkProcessor):
    """Handles Amazon SageMaker processing tasks for jobs using HuggingFace containers."""

    estimator_cls = HuggingFace

    def __init__(
        self,
        role: str,
        instance_count: Union[int, PipelineVariable],
        instance_type: Union[str, PipelineVariable],
        transformers_version: Optional[str] = None,
        tensorflow_version: Optional[str] = None,
        pytorch_version: Optional[str] = None,
        py_version: str = "py36",
        image_uri: Optional[Union[str, PipelineVariable]] = None,
        command: Optional[List[str]] = None,
        volume_size_in_gb: Union[int, PipelineVariable] = 30,
        volume_kms_key: Optional[Union[str, PipelineVariable]] = None,
        output_kms_key: Optional[Union[str, PipelineVariable]] = None,
        code_location: Optional[str] = None,
        max_runtime_in_seconds: Optional[Union[int, PipelineVariable]] = None,
        base_job_name: Optional[str] = None,
        sagemaker_session: Optional[Session] = None,
        env: Optional[Dict[str, Union[str, PipelineVariable]]] = None,
        tags: Optional[List[Dict[str, Union[str, PipelineVariable]]]] = None,
        network_config: Optional[NetworkConfig] = None,
    ):
        """This processor executes a Python script in a HuggingFace execution environment.

        Unless ``image_uri`` is specified, the environment is an Amazon-built Docker container
        that executes functions defined in the supplied ``code`` Python script.

        The arguments have the same meaning as in ``FrameworkProcessor``, with the following
        exceptions.

        Args:
            transformers_version (str): Transformers version you want to use for
                executing your model training code. Defaults to ``None``. Required unless
                ``image_uri`` is provided. The current supported version is ``4.4.2``.
            tensorflow_version (str): TensorFlow version you want to use for
                executing your model training code. Defaults to ``None``. Required unless
                ``pytorch_version`` is provided. The current supported version is ``1.6.0``.
            pytorch_version (str): PyTorch version you want to use for
                executing your model training code. Defaults to ``None``. Required unless
                ``tensorflow_version`` is provided. The current supported version is ``2.4.1``.
            py_version (str): Python version you want to use for executing your model training
                code. Defaults to ``None``. Required unless ``image_uri`` is provided.  If
                using PyTorch, the current supported version is ``py36``. If using TensorFlow,
                the current supported version is ``py37``.

        .. tip::

            You can find additional parameters for initializing this class at
            :class:`~sagemaker.processing.FrameworkProcessor`.
        """
        self.pytorch_version = pytorch_version
        self.tensorflow_version = tensorflow_version
        super().__init__(
            self.estimator_cls,
            transformers_version,
            role,
            instance_count,
            instance_type,
            py_version,
            image_uri,
            command,
            volume_size_in_gb,
            volume_kms_key,
            output_kms_key,
            code_location,
            max_runtime_in_seconds,
            base_job_name,
            sagemaker_session,
            env,
            tags,
            network_config,
        )

    def _create_estimator(
        self,
        entry_point="",
        source_dir=None,
        dependencies=None,
        git_config=None,
    ):
        """Override default estimator factory function for HuggingFace's different parameters

        HuggingFace estimators have 3 framework version parameters instead of one: The version for
        Transformers, PyTorch, and TensorFlow.
        """
        return self.estimator_cls(
            transformers_version=self.framework_version,
            tensorflow_version=self.tensorflow_version,
            pytorch_version=self.pytorch_version,
            py_version=self.py_version,
            entry_point=entry_point,
            source_dir=source_dir,
            dependencies=dependencies,
            git_config=git_config,
            code_location=self.code_location,
            enable_network_isolation=False,
            image_uri=self.image_uri,
            role=self.role,
            instance_count=self.instance_count,
            instance_type=self.instance_type,
            sagemaker_session=self.sagemaker_session,
            debugger_hook_config=False,
            disable_profiler=True,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""The pipeline context for workflow"""
from __future__ import absolute_import

import warnings
import inspect
from functools import wraps
from typing import Dict, Optional, Callable

from sagemaker.session import Session, SessionSettings
from sagemaker.local import LocalSession


class _StepArguments:
    """Step arguments entity for `Step`"""

    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, caller_name: str = None, func: Callable = None, *func_args, **func_kwargs):
        """Create a `_StepArguments`

        Args:
            caller_name (str): The name of the caller function which is intercepted by the
                PipelineSession to get the step arguments.
            func (Callable): The job class function that generates the step arguments used
                when creating the job ( fit() for a training job )
            *func_args: The args for func
            **func_kwargs: The kwargs for func
        """
        self.caller_name = caller_name
        self.func = func
        self.func_args = func_args
        self.func_kwargs = func_kwargs


class _JobStepArguments(_StepArguments):
    """Step arguments entity for job step types

    Job step types include: TrainingStep, ProcessingStep, TuningStep, TransformStep
    """

    def __init__(self, caller_name: str, args: dict):
        """Create a `_JobStepArguments`

        Args:
            caller_name (str): The name of the caller function which is intercepted by the
                PipelineSession to get the step arguments.
            args (dict): The arguments to be used for composing the SageMaker API request.
        """
        super(_JobStepArguments, self).__init__(caller_name)
        self.args = args


class _ModelStepArguments(_StepArguments):
    """Step arguments entity for `ModelStep`"""

    def __init__(self, model):
        """Create a `_ModelStepArguments`

        Args:
            model (Model or PipelineModel): A `sagemaker.model.Model`
                or `sagemaker.pipeline.PipelineModel` instance
        """
        super(_ModelStepArguments, self).__init__()
        self.model = model
        self.create_model_package_request = None
        self.create_model_request = None
        self.need_runtime_repack = set()
        self.runtime_repack_output_prefix = None


class _PipelineConfig:
    """Config object that associates a step with its containing pipeline

    Args:
        pipeline_name (str): pipeline name
        step_name (str): step name
        code_hash (str): a hash of the code artifact for the particular step
        config_hash (str): a hash of the config artifact for the particular step (Processing)
    """

    def __init__(self, pipeline_name, step_name, code_hash, config_hash):
        self.pipeline_name = pipeline_name
        self.step_name = step_name
        self.code_hash = code_hash
        self.config_hash = config_hash


class PipelineSession(Session):
    """Managing interactions with SageMaker APIs and AWS services needed under Pipeline Context

    This class inherits the SageMaker session, it provides convenient methods
    for manipulating entities and resources that Amazon SageMaker uses,
    such as training jobs, endpoints, and input datasets in S3. When composing
    SageMaker Model-Building Pipeline, PipelineSession is recommended over
    regular SageMaker Session
    """

    def __init__(
        self,
        boto_session=None,
        sagemaker_client=None,
        default_bucket=None,
        settings=SessionSettings(),
    ):
        """Initialize a ``PipelineSession``.

        Args:
            boto_session (boto3.session.Session): The underlying Boto3 session which AWS service
                calls are delegated to (default: None). If not provided, one is created with
                default AWS configuration chain.
            sagemaker_client (boto3.SageMaker.Client): Client which makes Amazon SageMaker service
                calls other than ``InvokeEndpoint`` (default: None). Estimators created using this
                ``Session`` use this client. If not provided, one will be created using this
                instance's ``boto_session``.
            default_bucket (str): The default Amazon S3 bucket to be used by this session.
                This will be created the next time an Amazon S3 bucket is needed (by calling
                :func:`default_bucket`).
                If not provided, a default bucket will be created based on the following format:
                "sagemaker-{region}-{aws-account-id}".
                Example: "sagemaker-my-custom-bucket".
            settings (sagemaker.session_settings.SessionSettings): Optional. Set of optional
                parameters to apply to the session.
        """
        super().__init__(
            boto_session=boto_session,
            sagemaker_client=sagemaker_client,
            default_bucket=default_bucket,
            settings=settings,
        )
        self._context = None

    @property
    def context(self):
        """Hold contextual information useful to the session"""
        return self._context

    @context.setter
    def context(self, value: Optional[_StepArguments] = None):
        self._context = value

    def _intercept_create_request(self, request: Dict, create, func_name: str = None):
        """This function intercepts the create job request

        Args:
            request (dict): the create job request
            create (functor): a functor calls the sagemaker client create method
            func_name (str): the name of the function needed intercepting
        """
        if func_name == self.create_model.__name__:
            self.context.create_model_request = request
            self.context.caller_name = func_name
        elif func_name == self.create_model_package_from_containers.__name__:
            self.context.create_model_package_request = request
            self.context.caller_name = func_name
        else:
            self.context = _JobStepArguments(func_name, request)

    def init_model_step_arguments(self, model):
        """Create a `_ModelStepArguments` (if not exist) as pipeline context

        Args:
            model (Model or PipelineModel): A `sagemaker.model.Model`
                or `sagemaker.pipeline.PipelineModel` instance
        """
        if not isinstance(self._context, _ModelStepArguments):
            self._context = _ModelStepArguments(model)


class LocalPipelineSession(LocalSession, PipelineSession):
    """Managing a session that executes Sagemaker pipelines and jobs locally in a pipeline context.

    This class inherits from the LocalSession and PipelineSession classes.
    When running Sagemaker pipelines locally, this class is preferred over LocalSession.
    """

    def __init__(
        self, boto_session=None, default_bucket=None, s3_endpoint_url=None, disable_local_code=False
    ):
        """Initialize a ``LocalPipelineSession``.

        Args:
            boto_session (boto3.session.Session): The underlying Boto3 session which AWS service
                calls are delegated to (default: None). If not provided, one is created with
                default AWS configuration chain.
            default_bucket (str): The default Amazon S3 bucket to be used by this session.
                This will be created the next time an Amazon S3 bucket is needed (by calling
                :func:`default_bucket`).
                If not provided, a default bucket will be created based on the following format:
                "sagemaker-{region}-{aws-account-id}".
                Example: "sagemaker-my-custom-bucket".
            s3_endpoint_url (str): Override the default endpoint URL for Amazon S3,
                if set (default: None).
            disable_local_code (bool): Set to True to override the default AWS configuration chain
                to disable the `local.local_code` setting, which may not be supported for some SDK
                features (default: False).
        """

        super().__init__(
            boto_session=boto_session,
            default_bucket=default_bucket,
            s3_endpoint_url=s3_endpoint_url,
            disable_local_code=disable_local_code,
        )


def runnable_by_pipeline(run_func):
    """A convenient Decorator

    This is a decorator designed to annotate, during pipeline session,
    the methods that downstream managed to
        1. preprocess user inputs, outputs, and configurations
        2. generate the create request
        3. start the job.
    For instance, `Processor.run`, `Estimator.fit`, or `Transformer.transform`.
    This decorator will essentially run 1, and capture the request shape from 2,
    then instead of starting a new job in 3, it will return request shape from 2
    to `sagemaker.workflow.steps.Step`. The request shape will be used to construct
    the arguments needed to compose that particular step as part of the pipeline.
    The job will be started during pipeline execution.
    """

    @wraps(run_func)
    def wrapper(*args, **kwargs):
        self_instance = args[0]
        if isinstance(self_instance.sagemaker_session, PipelineSession):
            run_func_params = inspect.signature(run_func).parameters
            arg_list = list(args)

            override_wait, override_logs = False, False
            for i, (arg_name, _) in enumerate(run_func_params.items()):
                if i >= len(arg_list):
                    break
                if arg_name == "wait":
                    override_wait = True
                    arg_list[i] = False
                elif arg_name == "logs":
                    override_logs = True
                    arg_list[i] = False

            args = tuple(arg_list)

            if not override_wait and "wait" in run_func_params.keys():
                kwargs["wait"] = False
            if not override_logs and "logs" in run_func_params.keys():
                kwargs["logs"] = False

            warnings.warn(
                "Running within a PipelineSession, there will be No Wait, "
                "No Logs, and No Job being started.",
                UserWarning,
            )
            if run_func.__name__ in ["register", "create"]:
                self_instance.sagemaker_session.init_model_step_arguments(self_instance)
                run_func(*args, **kwargs)
                context = self_instance.sagemaker_session.context
                self_instance.sagemaker_session.context = None
                return context

            return _StepArguments(retrieve_caller_name(self_instance), run_func, *args, **kwargs)

        return run_func(*args, **kwargs)

    return wrapper


def retrieve_caller_name(job_instance):
    """Convenience method for runnable_by_pipeline decorator

    This function takes an instance of a job class and maps it
    to the pipeline session function that creates the job request.

    Args:
        job_instance: A job class instance, one of the following
            imported types
    """

    from sagemaker.processing import Processor
    from sagemaker.estimator import EstimatorBase
    from sagemaker.transformer import Transformer
    from sagemaker.tuner import HyperparameterTuner
    from sagemaker.automl.automl import AutoML

    if isinstance(job_instance, Processor):
        return "process"
    if isinstance(job_instance, EstimatorBase):
        return "train"
    if isinstance(job_instance, Transformer):
        return "transform"
    if isinstance(job_instance, HyperparameterTuner):
        return "create_tuning_job"
    if isinstance(job_instance, AutoML):
        return "auto_ml"

    return None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""A class for AsyncInferenceConfig

Used for configuring async inference endpoint. Use AsyncInferenceConfig when deploying
the model to the async inference endpoints.
"""
from __future__ import print_function, absolute_import


class AsyncInferenceConfig(object):
    """Configuration object passed in when deploying models to Amazon SageMaker Endpoints.

    This object specifies configuration related to async endpoint. Use this configuration
    when trying to create async endpoint and make async inference
    """

    def __init__(
        self,
        output_path=None,
        max_concurrent_invocations_per_instance=None,
        kms_key_id=None,
        notification_config=None,
    ):
        """Initialize an AsyncInferenceConfig object for async inference configuration.

        Args:
            output_path (str): Optional. The Amazon S3 location that endpoints upload
                inference responses to. If no value is provided, Amazon SageMaker will
                use default Amazon S3 Async Inference output path. (Default: None)
            max_concurrent_invocations_per_instance (int): Optional. The maximum number of
                concurrent requests sent by the SageMaker client to the model container. If
                no value is provided, Amazon SageMaker will choose an optimal value for you.
                (Default: None)
            kms_key_id (str): Optional. The Amazon Web Services Key Management Service
                (Amazon Web Services KMS) key that Amazon SageMaker uses to encrypt the
                asynchronous inference output in Amazon S3. (Default: None)
            notification_config (dict): Optional. Specifies the configuration for notifications
                of inference results for asynchronous inference. Only one notification is generated
                per invocation request (Default: None):
                * success_topic (str): Amazon SNS topic to post a notification to when inference
                completes successfully. If no topic is provided, no notification is sent on success.
                The key in notification_config is 'SuccessTopic'.
                * error_topic (str): Amazon SNS topic to post a notification to when inference
                fails. If no topic is provided, no notification is sent on failure.
                The key in notification_config is 'ErrorTopic'.
        """
        self.output_path = output_path
        self.max_concurrent_invocations_per_instance = max_concurrent_invocations_per_instance
        self.kms_key_id = kms_key_id
        self.notification_config = notification_config

    def _to_request_dict(self):
        """Generates a request dictionary using the parameters provided to the class."""
        request_dict = {
            "OutputConfig": {
                "S3OutputPath": self.output_path,
            },
        }

        if self.max_concurrent_invocations_per_instance:
            request_dict["ClientConfig"] = {
                "MaxConcurrentInvocationsPerInstance": self.max_concurrent_invocations_per_instance
            }

        if self.kms_key_id:
            request_dict["OutputConfig"]["KmsKeyId"] = self.kms_key_id

        if self.notification_config:
            request_dict["OutputConfig"]["NotificationConfig"] = self.notification_config

        return request_dict



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Amazon SageMaker channel configurations for S3 data sources and file system data sources"""
from __future__ import absolute_import, print_function

from typing import Union, Optional, List
import attr

from sagemaker.workflow.entities import PipelineVariable

FILE_SYSTEM_TYPES = ["FSxLustre", "EFS"]
FILE_SYSTEM_ACCESS_MODES = ["ro", "rw"]


class TrainingInput(object):
    """Amazon SageMaker channel configurations for S3 data sources.

    Attributes:
        config (dict[str, dict]): A SageMaker ``DataSource`` referencing
            a SageMaker ``S3DataSource``.
    """

    def __init__(
        self,
        s3_data: Union[str, PipelineVariable],
        distribution: Optional[Union[str, PipelineVariable]] = None,
        compression: Optional[Union[str, PipelineVariable]] = None,
        content_type: Optional[Union[str, PipelineVariable]] = None,
        record_wrapping: Optional[Union[str, PipelineVariable]] = None,
        s3_data_type: Union[str, PipelineVariable] = "S3Prefix",
        instance_groups: Optional[List[Union[str, PipelineVariable]]] = None,
        input_mode: Optional[Union[str, PipelineVariable]] = None,
        attribute_names: Optional[List[Union[str, PipelineVariable]]] = None,
        target_attribute_name: Optional[Union[str, PipelineVariable]] = None,
        shuffle_config: Optional["ShuffleConfig"] = None,
    ):
        r"""Create a definition for input data used by an SageMaker training job.

        See AWS documentation on the ``CreateTrainingJob`` API for more details
        on the parameters.

        Args:
            s3_data (str or PipelineVariable): Defines the location of S3 data to train on.
            distribution (str or PipelineVariable): Valid values: ``'FullyReplicated'``,
                ``'ShardedByS3Key'`` (default: ``'FullyReplicated'``).
            compression (str or PipelineVariable): Valid values: ``'Gzip'``, ``None``
                (default: None). This is used only in Pipe input mode.
            content_type (str or PipelineVariable): MIME type of the input data
                (default: None).
            record_wrapping (str or PipelineVariable): Valid values: 'RecordIO'
                (default: None).
            s3_data_type (str or PipelineVariable): Valid values: ``'S3Prefix'``,
                ``'ManifestFile'``, ``'AugmentedManifestFile'``.
                If ``'S3Prefix'``, ``s3_data`` defines a prefix of s3 objects to train on.
                All objects with s3 keys beginning with ``s3_data`` will be used to train.
                If ``'ManifestFile'`` or ``'AugmentedManifestFile'``,
                then ``s3_data`` defines a
                single S3 manifest file or augmented manifest file respectively,
                listing the S3 data to train on. Both the ManifestFile and
                AugmentedManifestFile formats are described at `S3DataSource
                <https://docs.aws.amazon.com/sagemaker/latest/dg/API_S3DataSource.html>`_
                in the `Amazon SageMaker API reference`.
            instance_groups (list[str] or list[PipelineVariable]): Optional. A list of
                instance group names in string format that you specified while configuring
                a heterogeneous cluster using the :class:`sagemaker.instance_group.InstanceGroup`.
                S3 data will be sent to all instance groups in the specified list.
                For instructions on how to use InstanceGroup objects
                to configure a heterogeneous cluster
                through the SageMaker generic and framework estimator classes, see
                `Train Using a Heterogeneous Cluster
                <https://docs.aws.amazon.com/sagemaker/latest/dg/train-heterogeneous-cluster.html>`_
                in the *Amazon SageMaker developer guide*.
                (default: None)
            input_mode (str or PipelineVariable): Optional override for this channel's input mode
                (default: None). By default, channels will use the input mode defined on
                ``sagemaker.estimator.EstimatorBase.input_mode``, but they will ignore
                that setting if this parameter is set.

                    * None - Amazon SageMaker will use the input mode specified in the ``Estimator``
                    * 'File' - Amazon SageMaker copies the training dataset from the S3 location to
                        a local directory.
                    * 'Pipe' - Amazon SageMaker streams data directly from S3 to the container via
                        a Unix-named pipe.
                    * 'FastFile' - Amazon SageMaker streams data from S3 on demand instead of
                        downloading the entire dataset before training begins.

            attribute_names (list[str] or list[PipelineVariable]): A list of one or more attribute
                names to use that are found in a specified AugmentedManifestFile.
            target_attribute_name (str or PipelineVariable): The name of the attribute will be
                predicted (classified) in a SageMaker AutoML job. It is required if the input is
                for SageMaker AutoML job.
            shuffle_config (sagemaker.inputs.ShuffleConfig): If specified this configuration enables
                shuffling on this channel. See the SageMaker API documentation for more info:
                https://docs.aws.amazon.com/sagemaker/latest/dg/API_ShuffleConfig.html
        """
        self.config = {
            "DataSource": {"S3DataSource": {"S3DataType": s3_data_type, "S3Uri": s3_data}}
        }

        if not (target_attribute_name or distribution):
            distribution = "FullyReplicated"

        if distribution is not None:
            self.config["DataSource"]["S3DataSource"]["S3DataDistributionType"] = distribution

        if compression is not None:
            self.config["CompressionType"] = compression
        if content_type is not None:
            self.config["ContentType"] = content_type
        if record_wrapping is not None:
            self.config["RecordWrapperType"] = record_wrapping
        if instance_groups is not None:
            self.config["DataSource"]["S3DataSource"]["InstanceGroupNames"] = instance_groups
        if input_mode is not None:
            self.config["InputMode"] = input_mode
        if attribute_names is not None:
            self.config["DataSource"]["S3DataSource"]["AttributeNames"] = attribute_names
        if target_attribute_name is not None:
            self.config["TargetAttributeName"] = target_attribute_name
        if shuffle_config is not None:
            self.config["ShuffleConfig"] = {"Seed": shuffle_config.seed}


class ShuffleConfig(object):
    """For configuring channel shuffling using a seed.

    For more detail, see the AWS documentation:
    https://docs.aws.amazon.com/sagemaker/latest/dg/API_ShuffleConfig.html
    """

    def __init__(self, seed):
        """Create a ShuffleConfig.

        Args:
            seed (long): the long value used to seed the shuffled sequence.
        """
        self.seed = seed


@attr.s
class CreateModelInput(object):
    """A class containing parameters which can be used to create a SageMaker Model

    Parameters:
        instance_type (str): type or EC2 instance will be used for model deployment.
        accelerator_type (str): elastic inference accelerator type.
    """

    instance_type: str = attr.ib(default=None)
    accelerator_type: str = attr.ib(default=None)


@attr.s
class TransformInput(object):
    """Create a class containing all the parameters.

    It can be used when calling ``sagemaker.transformer.Transformer.transform()``
    """

    data: str = attr.ib()
    data_type: str = attr.ib(default="S3Prefix")
    content_type: str = attr.ib(default=None)
    compression_type: str = attr.ib(default=None)
    split_type: str = attr.ib(default=None)
    input_filter: str = attr.ib(default=None)
    output_filter: str = attr.ib(default=None)
    join_source: str = attr.ib(default=None)
    model_client_config: dict = attr.ib(default=None)
    batch_data_capture_config: dict = attr.ib(default=None)


class FileSystemInput(object):
    """Amazon SageMaker channel configurations for file system data sources.

    Attributes:
        config (dict[str, dict]): A Sagemaker File System ``DataSource``.
    """

    def __init__(
        self,
        file_system_id,
        file_system_type,
        directory_path,
        file_system_access_mode="ro",
        content_type=None,
    ):
        """Create a new file system input used by an SageMaker training job.

        Args:
            file_system_id (str): An Amazon file system ID starting with 'fs-'.
            file_system_type (str): The type of file system used for the input.
                Valid values: 'EFS', 'FSxLustre'.
            directory_path (str): Absolute or normalized path to the root directory (mount point) in
                the file system.
                Reference: https://docs.aws.amazon.com/efs/latest/ug/mounting-fs.html and
                https://docs.aws.amazon.com/fsx/latest/LustreGuide/mount-fs-auto-mount-onreboot.html
            file_system_access_mode (str): Permissions for read and write.
                Valid values: 'ro' or 'rw'. Defaults to 'ro'.
        """

        if file_system_type not in FILE_SYSTEM_TYPES:
            raise ValueError(
                "Unrecognized file system type: %s. Valid values: %s."
                % (file_system_type, ", ".join(FILE_SYSTEM_TYPES))
            )

        if file_system_access_mode not in FILE_SYSTEM_ACCESS_MODES:
            raise ValueError(
                "Unrecognized file system access mode: %s. Valid values: %s."
                % (file_system_access_mode, ", ".join(FILE_SYSTEM_ACCESS_MODES))
            )

        self.config = {
            "DataSource": {
                "FileSystemDataSource": {
                    "FileSystemId": file_system_id,
                    "FileSystemType": file_system_type,
                    "DirectoryPath": directory_path,
                    "FileSystemAccessMode": file_system_access_mode,
                }
            }
        }

        if content_type:
            self.config["ContentType"] = content_type


class BatchDataCaptureConfig(object):
    """Configuration object passed in when create a batch transform job.

    Specifies configuration related to batch transform job data capture for use with
    Amazon SageMaker Model Monitoring
    """

    def __init__(
        self,
        destination_s3_uri: str,
        kms_key_id: str = None,
        generate_inference_id: bool = None,
    ):
        """Create new BatchDataCaptureConfig

        Args:
            destination_s3_uri (str): S3 Location to store the captured data
            kms_key_id (str): The KMS key to use when writing to S3.
                KmsKeyId can be an ID of a KMS key, ARN of a KMS key, alias of a KMS key,
                or alias of a KMS key. The KmsKeyId is applied to all outputs.
                (default: None)
            generate_inference_id (bool): Flag to generate an inference id
                (default: None)
        """
        self.destination_s3_uri = destination_s3_uri
        self.kms_key_id = kms_key_id
        self.generate_inference_id = generate_inference_id

    def _to_request_dict(self):
        """Generates a request dictionary using the parameters provided to the class."""
        batch_data_capture_config = {
            "DestinationS3Uri": self.destination_s3_uri,
        }

        if self.kms_key_id is not None:
            batch_data_capture_config["KmsKeyId"] = self.kms_key_id
        if self.generate_inference_id is not None:
            batch_data_capture_config["GenerateInferenceId"] = self.generate_inference_id

        return batch_data_capture_config



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains code to query SageMaker lineage."""
from __future__ import absolute_import

from datetime import datetime
from enum import Enum
from typing import Optional, Union, List, Dict

from sagemaker.lineage._utils import get_resource_name_from_arn


class LineageEntityEnum(Enum):
    """Enum of lineage entities for use in a query filter."""

    TRIAL = "Trial"
    ACTION = "Action"
    ARTIFACT = "Artifact"
    CONTEXT = "Context"
    TRIAL_COMPONENT = "TrialComponent"


class LineageSourceEnum(Enum):
    """Enum of lineage types for use in a query filter."""

    CHECKPOINT = "Checkpoint"
    DATASET = "DataSet"
    ENDPOINT = "Endpoint"
    IMAGE = "Image"
    MODEL = "Model"
    MODEL_DATA = "ModelData"
    MODEL_DEPLOYMENT = "ModelDeployment"
    MODEL_GROUP = "ModelGroup"
    MODEL_REPLACE = "ModelReplaced"
    TENSORBOARD = "TensorBoard"
    TRAINING_JOB = "TrainingJob"
    APPROVAL = "Approval"
    PROCESSING_JOB = "ProcessingJob"
    TRANSFORM_JOB = "TransformJob"


class LineageQueryDirectionEnum(Enum):
    """Enum of query filter directions."""

    BOTH = "Both"
    ASCENDANTS = "Ascendants"
    DESCENDANTS = "Descendants"


class Edge:
    """A connecting edge for a lineage graph."""

    def __init__(
        self,
        source_arn: str,
        destination_arn: str,
        association_type: str,
    ):
        """Initialize ``Edge`` instance."""
        self.source_arn = source_arn
        self.destination_arn = destination_arn
        self.association_type = association_type

    def __hash__(self):
        """Define hash function for ``Edge``."""
        return hash(
            (
                "source_arn",
                self.source_arn,
                "destination_arn",
                self.destination_arn,
                "association_type",
                self.association_type,
            )
        )

    def __eq__(self, other):
        """Define equal function for ``Edge``."""
        return (
            self.association_type == other.association_type
            and self.source_arn == other.source_arn
            and self.destination_arn == other.destination_arn
        )


class Vertex:
    """A vertex for a lineage graph."""

    def __init__(
        self,
        arn: str,
        lineage_entity: str,
        lineage_source: str,
        sagemaker_session,
    ):
        """Initialize ``Vertex`` instance."""
        self.arn = arn
        self.lineage_entity = lineage_entity
        self.lineage_source = lineage_source
        self._session = sagemaker_session

    def __hash__(self):
        """Define hash function for ``Vertex``."""
        return hash(
            (
                "arn",
                self.arn,
                "lineage_entity",
                self.lineage_entity,
                "lineage_source",
                self.lineage_source,
            )
        )

    def __eq__(self, other):
        """Define equal function for ``Vertex``."""
        return (
            self.arn == other.arn
            and self.lineage_entity == other.lineage_entity
            and self.lineage_source == other.lineage_source
        )

    def to_lineage_object(self):
        """Convert the ``Vertex`` object to its corresponding lineage object.

        Returns:
            A ``Vertex`` object to its corresponding ``Artifact``,``Action``, ``Context``
            or ``TrialComponent`` object.
        """
        from sagemaker.lineage.context import Context, EndpointContext
        from sagemaker.lineage.action import Action
        from sagemaker.lineage.lineage_trial_component import LineageTrialComponent

        if self.lineage_entity == LineageEntityEnum.CONTEXT.value:
            resource_name = get_resource_name_from_arn(self.arn)
            if self.lineage_source == LineageSourceEnum.ENDPOINT.value:
                return EndpointContext.load(
                    context_name=resource_name, sagemaker_session=self._session
                )
            return Context.load(context_name=resource_name, sagemaker_session=self._session)

        if self.lineage_entity == LineageEntityEnum.ARTIFACT.value:
            return self._artifact_to_lineage_object()

        if self.lineage_entity == LineageEntityEnum.ACTION.value:
            return Action.load(action_name=self.arn.split("/")[1], sagemaker_session=self._session)

        if self.lineage_entity == LineageEntityEnum.TRIAL_COMPONENT.value:
            trial_component_name = get_resource_name_from_arn(self.arn)
            return LineageTrialComponent.load(
                trial_component_name=trial_component_name, sagemaker_session=self._session
            )
        raise ValueError("Vertex cannot be converted to a lineage object.")

    def _artifact_to_lineage_object(self):
        """Convert the ``Vertex`` object to its corresponding ``Artifact``."""
        from sagemaker.lineage.artifact import Artifact, ModelArtifact, ImageArtifact
        from sagemaker.lineage.artifact import DatasetArtifact

        if self.lineage_source == LineageSourceEnum.MODEL.value:
            return ModelArtifact.load(artifact_arn=self.arn, sagemaker_session=self._session)
        if self.lineage_source == LineageSourceEnum.DATASET.value:
            return DatasetArtifact.load(artifact_arn=self.arn, sagemaker_session=self._session)
        if self.lineage_source == LineageSourceEnum.IMAGE.value:
            return ImageArtifact.load(artifact_arn=self.arn, sagemaker_session=self._session)
        return Artifact.load(artifact_arn=self.arn, sagemaker_session=self._session)


class LineageQueryResult(object):
    """A wrapper around the results of a lineage query."""

    def __init__(
        self,
        edges: List[Edge] = None,
        vertices: List[Vertex] = None,
    ):
        """Init for LineageQueryResult.

        Args:
            edges (List[Edge]): The edges of the query result.
            vertices (List[Vertex]): The vertices of the query result.
        """
        self.edges = []
        self.vertices = []

        if edges is not None:
            self.edges = edges

        if vertices is not None:
            self.vertices = vertices


class LineageFilter(object):
    """A filter used in a lineage query."""

    def __init__(
        self,
        entities: Optional[List[Union[LineageEntityEnum, str]]] = None,
        sources: Optional[List[Union[LineageSourceEnum, str]]] = None,
        created_before: Optional[datetime] = None,
        created_after: Optional[datetime] = None,
        modified_before: Optional[datetime] = None,
        modified_after: Optional[datetime] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        """Initialize ``LineageFilter`` instance."""
        self.entities = entities
        self.sources = sources
        self.created_before = created_before
        self.created_after = created_after
        self.modified_before = modified_before
        self.modified_after = modified_after
        self.properties = properties

    def _to_request_dict(self):
        """Convert the lineage filter to its API representation."""
        filter_request = {}
        if self.sources:
            filter_request["Types"] = list(
                map(lambda x: x.value if isinstance(x, LineageSourceEnum) else x, self.sources)
            )
        if self.entities:
            filter_request["LineageTypes"] = list(
                map(lambda x: x.value if isinstance(x, LineageEntityEnum) else x, self.entities)
            )
        if self.created_before:
            filter_request["CreatedBefore"] = self.created_before
        if self.created_after:
            filter_request["CreatedAfter"] = self.created_after
        if self.modified_before:
            filter_request["ModifiedBefore"] = self.modified_before
        if self.modified_after:
            filter_request["ModifiedAfter"] = self.modified_after
        if self.properties:
            filter_request["Properties"] = self.properties
        return filter_request


class LineageQuery(object):
    """Creates an object used for performing lineage queries."""

    def __init__(self, sagemaker_session):
        """Initialize ``LineageQuery`` instance."""
        self._session = sagemaker_session

    def _get_edge(self, edge):
        """Convert lineage query API response to an Edge."""
        return Edge(
            source_arn=edge["SourceArn"],
            destination_arn=edge["DestinationArn"],
            association_type=edge["AssociationType"] if "AssociationType" in edge else None,
        )

    def _get_vertex(self, vertex):
        """Convert lineage query API response to a Vertex."""
        vertex_type = None
        if "Type" in vertex:
            vertex_type = vertex["Type"]
        return Vertex(
            arn=vertex["Arn"],
            lineage_source=vertex_type,
            lineage_entity=vertex["LineageType"],
            sagemaker_session=self._session,
        )

    def _convert_api_response(self, response) -> LineageQueryResult:
        """Convert the lineage query API response to its Python representation."""
        converted = LineageQueryResult()
        converted.edges = [self._get_edge(edge) for edge in response["Edges"]]
        converted.vertices = [self._get_vertex(vertex) for vertex in response["Vertices"]]

        edge_set = set()
        for edge in converted.edges:
            if edge in edge_set:
                converted.edges.remove(edge)
            edge_set.add(edge)

        vertex_set = set()
        for vertex in converted.vertices:
            if vertex in vertex_set:
                converted.vertices.remove(vertex)
            vertex_set.add(vertex)

        return converted

    def _collapse_cross_account_artifacts(self, query_response):
        """Collapse the duplicate vertices and edges for cross-account."""
        for edge in query_response.edges:
            if (
                "artifact" in edge.source_arn
                and "artifact" in edge.destination_arn
                and edge.source_arn.split("/")[1] == edge.destination_arn.split("/")[1]
                and edge.source_arn != edge.destination_arn
            ):
                edge_source_arn = edge.source_arn
                edge_destination_arn = edge.destination_arn
                self._update_cross_account_edge(
                    edges=query_response.edges,
                    arn=edge_source_arn,
                    duplicate_arn=edge_destination_arn,
                )
                self._update_cross_account_vertex(
                    query_response=query_response, duplicate_arn=edge_destination_arn
                )

        # remove the duplicate edges from cross account
        new_edge = [e for e in query_response.edges if not e.source_arn == e.destination_arn]
        query_response.edges = new_edge

        return query_response

    def _update_cross_account_edge(self, edges, arn, duplicate_arn):
        """Replace the duplicate arn with arn in edges list."""
        for idx, e in enumerate(edges):
            if e.destination_arn == duplicate_arn:
                edges[idx].destination_arn = arn
            elif e.source_arn == duplicate_arn:
                edges[idx].source_arn = arn

    def _update_cross_account_vertex(self, query_response, duplicate_arn):
        """Remove the vertex with duplicate arn in the vertices list."""
        query_response.vertices = [v for v in query_response.vertices if not v.arn == duplicate_arn]

    def query(
        self,
        start_arns: List[str],
        direction: LineageQueryDirectionEnum = LineageQueryDirectionEnum.BOTH,
        include_edges: bool = True,
        query_filter: LineageFilter = None,
        max_depth: int = 10,
    ) -> LineageQueryResult:
        """Perform a lineage query.

        Args:
            start_arns (List[str]): A list of ARNs that will be used as the starting point
                for the query.
            direction (LineageQueryDirectionEnum, optional): The direction of the query.
            include_edges (bool, optional): If true, return edges in addition to vertices.
            query_filter (LineageQueryFilter, optional): The query filter.

        Returns:
            LineageQueryResult: The lineage query result.
        """
        query_response = self._session.sagemaker_client.query_lineage(
            StartArns=start_arns,
            Direction=direction.value,
            IncludeEdges=include_edges,
            Filters=query_filter._to_request_dict() if query_filter else {},
            MaxDepth=max_depth,
        )
        query_response = self._convert_api_response(query_response)
        query_response = self._collapse_cross_account_artifacts(query_response)

        return query_response



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Contains API objects for SageMaker experiments."""
from __future__ import absolute_import

import enum
import numbers

from sagemaker.apiutils import _base_types


class TrialComponentMetricSummary(_base_types.ApiObject):
    """Summary model of a trial component.

    Attributes:
        metric_name (str): The name of the metric.
        source_arn (str):  The ARN of the source.
        time_stamp (datetime): Metric last updated value.
        max (float): The max value of the metric.
        min (float):  The min value of the metric.
        last (float):  The last value of the metric.
        count (float):  The number of samples used to generate the metric.
        avg (float):  The average value of the metric.
        std_dev (float):  The standard deviation of the metric.
    """

    metric_name = None
    source_arn = None
    time_stamp = None
    max = None
    min = None
    last = None
    count = None
    avg = None
    std_dev = None

    def __init__(self, metric_name=None, source_arn=None, **kwargs):
        super(TrialComponentMetricSummary, self).__init__(
            metric_name=metric_name, source_arn=source_arn, **kwargs
        )


class TrialComponentParameters(_base_types.ApiObject):
    """A dictionary of TrialComponentParameterValues"""

    @classmethod
    def from_boto(cls, boto_dict, **kwargs):
        """Converts a boto dict to a dictionary of TrialComponentParameterValues

        Args:
            boto_dict (dict): boto response dictionary.
            **kwargs:  Arbitrary keyword arguments.

        Returns:
            dict: Dictionary of parameter values.
        """
        return_map = {}
        for key, value in boto_dict.items():
            return_map[key] = value.get("NumberValue", value.get("StringValue", None))
        return return_map

    @classmethod
    def to_boto(cls, parameters):
        """Converts TrialComponentParameters to dict.

        Args:
            parameters (TrialComponentParameters): Dictionary to convert.

        Returns:
            dict: Dictionary of trial component parameters in boto format.
        """
        boto_map = {}
        for key, value in parameters.items():
            if isinstance(value, numbers.Number):
                boto_map[key] = {"NumberValue": value}
            else:
                boto_map[key] = {"StringValue": str(value)}
        return boto_map


class TrialComponentArtifact(_base_types.ApiObject):
    """Trial component artifact.

    Attributes:
        value (str): The artifact value.
        media_type (str): The media type.
    """

    value = None
    media_type = None

    def __init__(self, value=None, media_type=None, **kwargs):
        super(TrialComponentArtifact, self).__init__(value=value, media_type=media_type, **kwargs)


class _TrialComponentStatusType(enum.Enum):
    """The type of trial component status"""

    InProgress = "InProgress"
    Completed = "Completed"
    Failed = "Failed"


class TrialComponentStatus(_base_types.ApiObject):
    """Status of the trial component.

    Attributes:
        primary_status (str): The status of a trial component.
        message (str): Status message.
    """

    primary_status = None
    message = None

    def __init__(self, primary_status=None, message=None, **kwargs):
        super(TrialComponentStatus, self).__init__(
            primary_status=primary_status, message=message, **kwargs
        )


class TrialComponentSummary(_base_types.ApiObject):
    """Summary model of a trial component.

    Attributes:
        trial_component_name (str): Name of trial component.
        trial_component_arn (str): ARN of the trial component.
        display_name (str): Friendly display name in UI.
        source_arn (str): ARN of the trial component source.
        status (str): Status.
        start_time (datetime): Start time.
        end_time (datetime): End time.
        creation_time (datetime): Creation time.
        created_by (str): Created by.
        last_modified_time (datetime): Date last modified.
        last_modified_by (datetime): User last modified.
    """

    _custom_boto_types = {
        "status": (TrialComponentStatus, False),
    }
    trial_component_name = None
    trial_component_arn = None
    display_name = None
    source_arn = None
    status = None
    start_time = None
    end_time = None
    creation_time = None
    created_by = None
    last_modified_time = None
    last_modified_by = None


class TrialComponentSource(_base_types.ApiObject):
    """Trial Component Source

    Attributes:
        source_arn (str): The ARN of the source.
    """

    source_arn = None

    def __init__(self, source_arn=None, **kwargs):
        super(TrialComponentSource, self).__init__(source_arn=source_arn, **kwargs)


class Parent(_base_types.ApiObject):
    """The trial/experiment/run that a trial component is associated with.

    Attributes:
        trial_name (str): Name of the trial.
        experiment_name (str): Name of the experiment.
        run_name (str): Name of the run.
    """

    trial_name = None
    experiment_name = None
    run_name = None


class TrialComponentSearchResult(_base_types.ApiObject):
    """Summary model of an Trial Component search result.

    Attributes:
        trial_component_arn (str): ARN of the trial component.
        trial_component_name (str): Name of the trial component.
        display_name (str): Display name of the trial component for UI display.
        source (dict): The source of the trial component.
        status (dict): The status of the trial component.
        start_time (datetime): Start time.
        end_time (datetime): End time.
        creation_time (datetime): Creation time.
        created_by (str): Created by.
        last_modified_time (datetime): Date last modified.
        last_modified_by (datetime): User last modified.
        parameters (dict): The hyperparameters of the component.
        input_artifacts (dict): The input artifacts of the component.
        output_artifacts (dict): The output artifacts of the component.
        metrics (list): The metrics for the component.
        source_detail (dict): The source of the trial component.
        tags (list): The list of tags that are associated with the trial component.
        parents (list[Parent]): The parent of trial component.
    """

    _custom_boto_types = {
        "parents": (Parent, True),  # parents is a collection (list) of Parent objects
    }
    trial_component_arn = None
    trial_component_name = None
    display_name = None
    source = None
    status = None
    start_time = None
    end_time = None
    creation_time = None
    created_by = None
    last_modified_time = None
    last_modified_by = None
    parameters = None
    input_artifacts = None
    output_artifacts = None
    metrics = None
    source_detail = None
    tags = None
    parents = None


class TrialSummary(_base_types.ApiObject):
    """Summary model of a trial.

    Attributes:
        trial_arn (str): The ARN of the trial.
        trial_name (str): The name of the trial.
        creation_time (datetime):  When the trial was created.
        last_modified_time (datetime): When the trial was last modified.
    """

    trial_arn = None
    trial_name = None
    creation_time = None
    last_modified_time = None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Configuration for the SageMaker Training Compiler."""
from __future__ import absolute_import
import logging
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from sagemaker.training_compiler.config import TrainingCompilerConfig as BaseConfig

logger = logging.getLogger(__name__)


class TrainingCompilerConfig(BaseConfig):
    """The SageMaker Training Compiler configuration class."""

    SUPPORTED_INSTANCE_CLASS_PREFIXES = ["p3", "p3dn", "g4dn", "p4d", "g5"]
    MIN_SUPPORTED_VERSION = "2.9"

    def __init__(self, enabled=True, debug=False):
        """This class initializes a ``TrainingCompilerConfig`` instance.

        `Amazon SageMaker Training Compiler
        <https://docs.aws.amazon.com/sagemaker/latest/dg/training-compiler.html>`_
        is a feature of SageMaker Training
        and speeds up training jobs by optimizing model execution graphs.

        You can compile TensorFlow models
        by passing the object of this configuration class to the ``compiler_config``
        parameter of the :class:`~sagemaker.tensorflow.TensorFlow`
        estimator.

        Args:
            enabled (bool): Optional. Switch to enable SageMaker Training Compiler.
                The default is ``True``.
            debug (bool): Optional. Whether to dump detailed logs for debugging.
                This comes with a potential performance slowdown.
                The default is ``False``.

        **Example**: The following code shows the basic usage of the
        :class:`sagemaker.tensorflow.TrainingCompilerConfig()` class
        to run a TensorFlow training job with the compiler.

        .. code-block:: python

            from sagemaker.tensorflow import TensorFlow, TrainingCompilerConfig

            tensorflow_estimator=TensorFlow(
                ...
                compiler_config=TrainingCompilerConfig()
            )

        .. seealso::

            For more information about how to enable SageMaker Training Compiler
            for various training settings such as using TensorFlow-based models,
            PyTorch-based models, and distributed training,
            see `Enable SageMaker Training Compiler
            <https://docs.aws.amazon.com/sagemaker/latest/dg/training-compiler-enable.html>`_
            in the `Amazon SageMaker Training Compiler developer guide
            <https://docs.aws.amazon.com/sagemaker/latest/dg/training-compiler.html>`_.

        """

        super(TrainingCompilerConfig, self).__init__(enabled=enabled, debug=debug)

    @classmethod
    def validate(cls, estimator):
        """Checks if SageMaker Training Compiler is configured correctly.

        Args:
            estimator (str): A estimator object
                If SageMaker Training Compiler is enabled, it will validate whether
                the estimator is configured to be compatible with Training Compiler.

        Raises:
            ValueError: Raised if the requested configuration is not compatible
                        with SageMaker Training Compiler.
        """

        super(TrainingCompilerConfig, cls).validate(estimator)

        if estimator.framework_version:
            if Version(estimator.framework_version) in SpecifierSet(
                f"< {cls.MIN_SUPPORTED_VERSION}"
            ):
                error_helper_string = (
                    "SageMaker Training Compiler only supports TensorFlow version "
                    ">= {} but received {}"
                )
                error_helper_string = error_helper_string.format(
                    cls.MIN_SUPPORTED_VERSION, estimator.framework_version
                )
                raise ValueError(error_helper_string)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""The step definitions for workflow."""
from __future__ import absolute_import

import warnings
from typing import List, Union, Optional

import attr

from sagemaker.estimator import EstimatorBase
from sagemaker.model import Model
from sagemaker import PipelineModel
from sagemaker.predictor import Predictor
from sagemaker.transformer import Transformer
from sagemaker.workflow.entities import RequestType
from sagemaker.workflow.steps import Step, CreateModelStep, TransformStep
from sagemaker.workflow._utils import _RegisterModelStep, _RepackModelStep
from sagemaker.workflow.retry import RetryPolicy
from sagemaker.utils import update_container_with_inference_params


@attr.s
class StepCollection:
    """A wrapper of pipeline steps for workflow.

    Attributes:
        name (str): The name of the `StepCollection`.
        steps (List[Step]): A list of steps.
    """

    name: str = attr.ib()
    steps: List[Step] = attr.ib(factory=list)

    def request_dicts(self) -> List[RequestType]:
        """Get the request structure for workflow service calls."""
        return [step.to_request() for step in self.steps]

    @property
    def properties(self):
        """The properties of the particular `StepCollection`."""
        if not self.steps:
            return None
        return self.steps[-1].properties


class RegisterModel(StepCollection):  # pragma: no cover
    """Register Model step collection for workflow."""

    _REGISTER_MODEL_NAME_BASE = "RegisterModel"
    _REPACK_MODEL_NAME_BASE = "RepackModel"

    def __init__(
        self,
        name: str,
        content_types,
        response_types,
        inference_instances=None,
        transform_instances=None,
        estimator: EstimatorBase = None,
        model_data=None,
        depends_on: Optional[List[Union[str, Step, StepCollection]]] = None,
        repack_model_step_retry_policies: List[RetryPolicy] = None,
        register_model_step_retry_policies: List[RetryPolicy] = None,
        model_package_group_name=None,
        model_metrics=None,
        approval_status=None,
        image_uri=None,
        compile_model_family=None,
        display_name=None,
        description=None,
        tags=None,
        model: Union[Model, PipelineModel] = None,
        drift_check_baselines=None,
        customer_metadata_properties=None,
        domain=None,
        sample_payload_url=None,
        task=None,
        framework=None,
        framework_version=None,
        nearest_model_name=None,
        data_input_configuration=None,
        **kwargs,
    ):
        """Construct steps `_RepackModelStep` and `_RegisterModelStep` based on the estimator.

        Args:
            name (str): The name of the training step.
            estimator: The estimator instance.
            model_data: The S3 uri to the model data from training.
            content_types (list): The supported MIME types for the input data (default: None).
            response_types (list): The supported MIME types for the output data (default: None).
            inference_instances (list): A list of the instance types that are used to
                generate inferences in real-time (default: None).
            transform_instances (list): A list of the instance types on which a transformation
                job can be run or on which an endpoint can be deployed (default: None).
            depends_on (List[Union[str, Step, StepCollection]]): The list of `Step`/`StepCollection`
                names or `Step` instances or `StepCollection` instances that the first step
                in the collection depends on (default: None).
            repack_model_step_retry_policies (List[RetryPolicy]): The list of retry policies
                for the repack model step
            register_model_step_retry_policies (List[RetryPolicy]): The list of retry policies
                for register model step
            model_package_group_name (str): The Model Package Group name or Arn, exclusive to
                `model_package_name`, using `model_package_group_name` makes the Model Package
                versioned (default: None).
            model_metrics (ModelMetrics): ModelMetrics object (default: None).
            approval_status (str): Model Approval Status, values can be "Approved", "Rejected",
                or "PendingManualApproval" (default: "PendingManualApproval").
            image_uri (str): The container image uri for Model Package, if not specified,
                Estimator's training container image is used (default: None).
            compile_model_family (str): The instance family for the compiled model. If
                specified, a compiled model is used (default: None).
            description (str): Model Package description (default: None).
            tags (List[dict[str, str]]): The list of tags to attach to the model package group. Note
                that tags will only be applied to newly created model package groups; if the
                name of an existing group is passed to "model_package_group_name",
                tags will not be applied.
            model (object or Model): A PipelineModel object that comprises a list of models
                which gets executed as a serial inference pipeline or a Model object.
            drift_check_baselines (DriftCheckBaselines): DriftCheckBaselines object (default: None).
            customer_metadata_properties (dict[str, str]): A dictionary of key-value paired
                metadata properties (default: None).
            domain (str): Domain values can be "COMPUTER_VISION", "NATURAL_LANGUAGE_PROCESSING",
                "MACHINE_LEARNING" (default: None).
            sample_payload_url (str): The S3 path where the sample payload is stored
                (default: None).
            task (str): Task values which are supported by Inference Recommender are "FILL_MASK",
                "IMAGE_CLASSIFICATION", "OBJECT_DETECTION", "TEXT_GENERATION", "IMAGE_SEGMENTATION",
                "CLASSIFICATION", "REGRESSION", "OTHER" (default: None).
            framework (str): Machine learning framework of the model package container image
                (default: None).
            framework_version (str): Framework version of the Model Package Container Image
                (default: None).
            nearest_model_name (str): Name of a pre-trained machine learning benchmarked by
                Amazon SageMaker Inference Recommender (default: None).
            data_input_configuration (str): Input object for the model (default: None).

            **kwargs: additional arguments to `create_model`.
        """
        self.name = name
        steps: List[Step] = []
        repack_model = False
        self.model_list = None
        self.container_def_list = None
        subnets = None
        security_group_ids = None

        if estimator is not None:
            subnets = estimator.subnets
            security_group_ids = estimator.security_group_ids
        elif model is not None and model.vpc_config is not None:
            subnets = model.vpc_config["Subnets"]
            security_group_ids = model.vpc_config["SecurityGroupIds"]

        if "entry_point" in kwargs:
            repack_model = True
            entry_point = kwargs.pop("entry_point", None)
            source_dir = kwargs.pop("source_dir", None)
            dependencies = kwargs.pop("dependencies", None)
            kwargs = dict(**kwargs, output_kms_key=kwargs.pop("model_kms_key", None))

            repack_model_step = _RepackModelStep(
                name="{}-{}".format(self.name, self._REPACK_MODEL_NAME_BASE),
                depends_on=depends_on,
                retry_policies=repack_model_step_retry_policies,
                sagemaker_session=estimator.sagemaker_session,
                role=estimator.role,
                model_data=model_data,
                entry_point=entry_point,
                source_dir=source_dir,
                dependencies=dependencies,
                tags=tags,
                subnets=subnets,
                security_group_ids=security_group_ids,
                description=description,
                display_name=display_name,
                **kwargs,
            )
            steps.append(repack_model_step)
            model_data = repack_model_step.properties.ModelArtifacts.S3ModelArtifacts

            # remove kwargs consumed by model repacking step
            kwargs.pop("output_kms_key", None)

        elif model is not None:
            if isinstance(model, PipelineModel):
                self.model_list = model.models
            elif isinstance(model, Model):
                self.model_list = [model]

            for model_entity in self.model_list:
                if estimator is not None:
                    sagemaker_session = estimator.sagemaker_session
                    role = estimator.role
                else:
                    sagemaker_session = model_entity.sagemaker_session
                    role = model_entity.role
                if hasattr(model_entity, "entry_point") and model_entity.entry_point is not None:
                    repack_model = True
                    entry_point = model_entity.entry_point
                    source_dir = model_entity.source_dir
                    dependencies = model_entity.dependencies
                    kwargs = dict(**kwargs, output_kms_key=model_entity.model_kms_key)
                    model_name = model_entity.name or model_entity._framework_name

                    repack_model_step = _RepackModelStep(
                        name="{}-{}".format(model_name, self._REPACK_MODEL_NAME_BASE),
                        depends_on=depends_on,
                        retry_policies=repack_model_step_retry_policies,
                        sagemaker_session=sagemaker_session,
                        role=role,
                        model_data=model_entity.model_data,
                        entry_point=entry_point,
                        source_dir=source_dir,
                        dependencies=dependencies,
                        tags=tags,
                        subnets=subnets,
                        security_group_ids=security_group_ids,
                        description=description,
                        display_name=display_name,
                        **kwargs,
                    )
                    steps.append(repack_model_step)
                    model_entity.model_data = (
                        repack_model_step.properties.ModelArtifacts.S3ModelArtifacts
                    )

                    # remove kwargs consumed by model repacking step
                    kwargs.pop("output_kms_key", None)

            if isinstance(model, PipelineModel):
                self.container_def_list = model.pipeline_container_def(
                    inference_instances[0] if inference_instances else None
                )
            elif isinstance(model, Model):
                self.container_def_list = [
                    model.prepare_container_def(
                        inference_instances[0] if inference_instances else None
                    )
                ]

            self.container_def_list = update_container_with_inference_params(
                framework=framework,
                framework_version=framework_version,
                nearest_model_name=nearest_model_name,
                data_input_configuration=data_input_configuration,
                container_list=self.container_def_list,
            )

        register_model_step = _RegisterModelStep(
            name="{}-{}".format(self.name, self._REGISTER_MODEL_NAME_BASE),
            estimator=estimator,
            model_data=model_data,
            content_types=content_types,
            response_types=response_types,
            inference_instances=inference_instances,
            transform_instances=transform_instances,
            model_package_group_name=model_package_group_name,
            model_metrics=model_metrics,
            drift_check_baselines=drift_check_baselines,
            approval_status=approval_status,
            image_uri=image_uri,
            compile_model_family=compile_model_family,
            description=description,
            display_name=display_name,
            tags=tags,
            container_def_list=self.container_def_list,
            retry_policies=register_model_step_retry_policies,
            customer_metadata_properties=customer_metadata_properties,
            domain=domain,
            sample_payload_url=sample_payload_url,
            task=task,
            **kwargs,
        )
        if not repack_model:
            register_model_step.add_depends_on(depends_on)

        steps.append(register_model_step)
        self.steps = steps

        warnings.warn(
            (
                "We are deprecating the use of RegisterModel. "
                "Please use the ModelStep instead. For more, see: "
                "https://sagemaker.readthedocs.io/en/stable/"
                "amazon_sagemaker_model_building_pipeline.html#model-step"
            ),
            DeprecationWarning,
        )


class EstimatorTransformer(StepCollection):
    """Creates a Transformer step collection for workflow."""

    def __init__(
        self,
        name: str,
        estimator: EstimatorBase,
        model_data,
        model_inputs,
        instance_count,
        instance_type,
        transform_inputs,
        description: str = None,
        display_name: str = None,
        # model arguments
        image_uri=None,
        predictor_cls=None,
        env=None,
        # transformer arguments
        strategy=None,
        assemble_with=None,
        output_path=None,
        output_kms_key=None,
        accept=None,
        max_concurrent_transforms=None,
        max_payload=None,
        tags=None,
        volume_kms_key=None,
        depends_on: Optional[List[Union[str, Step, StepCollection]]] = None,
        # step retry policies
        repack_model_step_retry_policies: List[RetryPolicy] = None,
        model_step_retry_policies: List[RetryPolicy] = None,
        transform_step_retry_policies: List[RetryPolicy] = None,
        **kwargs,
    ):
        """Construct steps required for a Transformer step collection:

        An estimator-centric step collection. It models what happens in workflows
        when invoking the `transform()` method on an estimator instance:
        First, if custom
        model artifacts are required, a `_RepackModelStep` is included.
        Second, a
        `CreateModelStep` with the model data passed in from a training step or other
        training job output.
        Finally, a `TransformerStep`.

        If repacking
        the model artifacts is not necessary, only the CreateModelStep and TransformerStep
        are in the step collection.

        Args:
            name (str): The name of the Transform Step.
            estimator: The estimator instance.
            instance_count (int): The number of EC2 instances to use.
            instance_type (str): The type of EC2 instance to use.
            strategy (str): The strategy used to decide how to batch records in
                a single request (default: None). Valid values: 'MultiRecord'
                and 'SingleRecord'.
            assemble_with (str): How the output is assembled (default: None).
                Valid values: 'Line' or 'None'.
            output_path (str): The S3 location for saving the transform result. If
                not specified, results are stored to a default bucket.
            output_kms_key (str): Optional. A KMS key ID for encrypting the
                transform output (default: None).
            accept (str): The accept header passed by the client to
                the inference endpoint. If it is supported by the endpoint,
                it will be the format of the batch transform output.
            env (dict): The Environment variables to be set for use during the
                transform job (default: None).
            depends_on (List[Union[str, Step, StepCollection]]): The list of `Step`/`StepCollection`
                names or `Step` instances or `StepCollection` instances that the first step
                in the collection depends on (default: None).
            repack_model_step_retry_policies (List[RetryPolicy]): The list of retry policies
                for the repack model step
            model_step_retry_policies (List[RetryPolicy]): The list of retry policies for
                model step
            transform_step_retry_policies (List[RetryPolicy]): The list of retry policies for
                transform step
        """
        self.name = name
        steps = []
        if "entry_point" in kwargs:
            entry_point = kwargs.get("entry_point", None)
            source_dir = kwargs.get("source_dir", None)
            dependencies = kwargs.get("dependencies", None)
            repack_model_step = _RepackModelStep(
                name=f"{name}RepackModel",
                depends_on=depends_on,
                retry_policies=repack_model_step_retry_policies,
                sagemaker_session=estimator.sagemaker_session,
                role=estimator.role,
                model_data=model_data,
                entry_point=entry_point,
                source_dir=source_dir,
                dependencies=dependencies,
                tags=tags,
                subnets=estimator.subnets,
                security_group_ids=estimator.security_group_ids,
                description=description,
                display_name=display_name,
            )
            steps.append(repack_model_step)
            model_data = repack_model_step.properties.ModelArtifacts.S3ModelArtifacts

        def predict_wrapper(endpoint, session):
            return Predictor(endpoint, session)

        predictor_cls = predictor_cls or predict_wrapper

        model = Model(
            image_uri=image_uri or estimator.training_image_uri(),
            model_data=model_data,
            predictor_cls=predictor_cls,
            vpc_config=None,
            sagemaker_session=estimator.sagemaker_session,
            role=estimator.role,
            env=kwargs.get("env", None),
            name=kwargs.get("name", None),
            enable_network_isolation=kwargs.get("enable_network_isolation", None),
            model_kms_key=kwargs.get("model_kms_key", None),
            image_config=kwargs.get("image_config", None),
        )
        model_step = CreateModelStep(
            name=f"{name}CreateModelStep",
            model=model,
            inputs=model_inputs,
            description=description,
            display_name=display_name,
            retry_policies=model_step_retry_policies,
        )
        if "entry_point" not in kwargs and depends_on:
            # if the CreateModelStep is the first step in the collection
            model_step.add_depends_on(depends_on)
        steps.append(model_step)

        transformer = Transformer(
            model_name=model_step.properties.ModelName,
            instance_count=instance_count,
            instance_type=instance_type,
            strategy=strategy,
            assemble_with=assemble_with,
            output_path=output_path,
            output_kms_key=output_kms_key,
            accept=accept,
            max_concurrent_transforms=max_concurrent_transforms,
            max_payload=max_payload,
            env=env,
            tags=tags,
            base_transform_job_name=name,
            volume_kms_key=volume_kms_key,
            sagemaker_session=estimator.sagemaker_session,
        )
        transform_step = TransformStep(
            name=f"{name}TransformStep",
            transformer=transformer,
            inputs=transform_inputs,
            description=description,
            display_name=display_name,
            retry_policies=transform_step_retry_policies,
        )
        steps.append(transform_step)

        self.steps = steps



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Custom exception classes for Sagemaker SDK"""
from __future__ import absolute_import


class UnexpectedStatusException(ValueError):
    """Raised when resource status is not expected and thus not allowed for further execution"""

    def __init__(self, message, allowed_statuses, actual_status):
        self.allowed_statuses = allowed_statuses
        self.actual_status = actual_status
        super(UnexpectedStatusException, self).__init__(message)


class CapacityError(UnexpectedStatusException):
    """Raised when resource status is not expected and fails with a reason of CapacityError"""


class AsyncInferenceError(Exception):
    """The base exception class for Async Inference exceptions."""

    fmt = "An unspecified error occurred"

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class ObjectNotExistedError(AsyncInferenceError):
    """Raised when Amazon S3 object not exist in the given path"""

    fmt = "Object not exist at {output_path}. {message}"

    def __init__(self, message, output_path):
        super().__init__(message=message, output_path=output_path)


class PollingTimeoutError(AsyncInferenceError):
    """Raised when wait longer than expected and no result object in Amazon S3 bucket yet"""

    fmt = "No result at {output_path} after polling for {seconds} seconds. {message}"

    def __init__(self, message, output_path, seconds):
        super().__init__(message=message, output_path=output_path, seconds=seconds)


class UnexpectedClientError(AsyncInferenceError):
    """Raised when ClientError's error code is not expected"""

    fmt = "Encountered unexpected client error: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class AutoMLStepInvalidModeError(Exception):
    """Raised when the automl mode passed into AutoMLStep in invalid"""

    fmt = (
        "Mode in AutoMLJobConfig must be defined for AutoMLStep. "
        "AutoMLStep currently only supports ENSEMBLING mode"
    )

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs
